"""Serial protocol to perform async I/O with the Insteon Modem."""

import asyncio
from enum import Enum
import logging
from queue import SimpleQueue
from typing import Union

from ..address import Address
from ..constants import AckNak
from ..utils import log_error, publish_topic
from .command_to_msg import register_command_handlers
from .messages.inbound import create
from .messages.outbound import outbound_write_manager, register_outbound_handlers
from .msg_to_topic import convert_to_topic

_LOGGER = logging.getLogger(__name__)
_LOGGER_PYINSTEON = logging.getLogger("pyinsteon")
_LOGGER_MSG = logging.getLogger("pyinsteon.messages")
MAX_RECONNECT_WAIT_TIME = 300


def _get_addresses_in_msg(msg):
    """Return a list of addresses included in a message."""
    addresses = []
    addr_fields = ["address", "target"]
    for fld in addr_fields:
        try:
            addr = getattr(msg, fld)
        except AttributeError:
            continue
        else:
            if addr not in addresses:
                addresses.append(addr)
    return addresses


# pylint: disable=broad-except
async def _publish_message(msg):
    """Convert an inbound message to a topic and publish to listeners."""
    _LOGGER_MSG.debug("RX: %s", repr(msg))
    if (_LOGGER_MSG.level == 0 or _LOGGER_MSG.level > logging.DEBUG) and (
        _LOGGER_PYINSTEON.level == 0 or _LOGGER_PYINSTEON.level > logging.DEBUG
    ):
        for addr in _get_addresses_in_msg(msg):
            logger = logging.getLogger(f"pyinsteon.{Address(addr).id}")
            logger.debug("RX: %s", repr(msg))
    topic = None
    kwargs = {}
    try:
        for topic, kwargs in convert_to_topic(msg):
            publish_topic(topic, **kwargs)
    except ValueError:
        # No topic was found for this message
        _LOGGER.debug("No topic found for message %r", msg)
    except Exception as ex:
        log_error(msg, ex, topic=topic, kwargs=kwargs)


class TransportStatus(Enum):
    """Status of the transport."""

    CLOSED = 0
    LOST = 1
    PAUSED = 2
    OPEN = 3


class Protocol(asyncio.Protocol):
    """Serial protocol to perform async I/O with the PLM."""

    def __init__(self, connect_method, *args, **kwargs):
        """Init the SerialProtocol class."""
        super().__init__(*args, **kwargs)
        self._transport = None
        self._message_queue = asyncio.PriorityQueue()
        self._last_message = SimpleQueue()
        self._buffer = bytearray()
        self._should_reconnect = True
        self._connect_method = connect_method
        self._writer_task = None
        self._writer_lock = asyncio.Lock()
        outbound_write_manager.protocol_write = self.write
        register_outbound_handlers()
        register_command_handlers()

    @property
    def connected(self) -> bool:
        """Return true if the transport is connected."""
        return not self._transport.is_closing() if self._transport else False

    @property
    def message_queue(self):
        """Return the queue of messages to write to the transport."""
        return self._message_queue

    @property
    def transport(self):
        """Return the transport."""
        return self._transport

    def connection_made(self, transport):
        """Run when a connection to the transport has been made."""
        self._transport = transport
        publish_topic("connection.made")

    def data_received(self, data):
        """Receive data from the serial transport."""
        self._buffer.extend(data)
        while True:
            last_buffer = self._buffer
            try:
                msg, self._buffer = create(self._buffer)
                if isinstance(self._buffer, bytes):
                    _LOGGER.warning("Buffer became bytes: %s", data.hex())
                    self._buffer = bytearray(self._buffer)
            except (ValueError, IndexError) as ex:
                _LOGGER.debug("Invalid message data: %s", self._buffer.hex())
                _LOGGER.debug("%s: %s", type(ex), str(ex))
                self._buffer = self._buffer[1:]
                msg = None

            # Sometimes the modem only responds with NAK and not the original message
            if (
                not msg
                and last_buffer
                and last_buffer[-1] == AckNak.NAK
                and not self._last_message.empty()
            ):
                last_msg = self._last_message.get()
                last_msg_nak = bytearray(bytes(last_msg))
                last_msg_nak.extend(bytes([0x15]))
                msg, _ = create(last_msg_nak)
            if msg:
                asyncio.create_task(_publish_message(msg))
                msg = None

            if not self._buffer or last_buffer == self._buffer:
                break

    def connection_lost(self, exc: Union[asyncio.Task, Exception]):
        """Notify listeners that the serial connection is lost."""
        _LOGGER.debug("Connection lost called")
        _LOGGER.debug("Should reconnect: %s", self._should_reconnect)
        if exc:
            if hasattr(exc, "exception"):
                log_msg = str(exc.exception())
            else:
                log_msg = str(exc)
            _LOGGER.warning("pyinsteon transport exception: %s", log_msg)
        if self._should_reconnect:
            asyncio.create_task(self.async_connect())
        else:
            asyncio.create_task(self._stop_writer())

    async def async_connect(self, retry=True):
        """Connect to the transport asynchronously."""
        wait_time = 0.1
        await asyncio.sleep(0.5)  # Give everything time to settle
        while not self.connected:
            _LOGGER.debug("Attempting to connect to modem")
            self._transport = await self._connect_method(protocol=self)
            if self._transport is None and not retry:
                publish_topic("connection.failed")
                raise ConnectionError("Modem did not respond to connection request")
            await asyncio.sleep(0.1)  # Let the transport finish connecting
            if not self.connected and retry:
                await asyncio.sleep(wait_time)
                wait_time = min(MAX_RECONNECT_WAIT_TIME, 1.5 * wait_time)
        if not self._writer_lock.locked():
            self._start_writer()

        _LOGGER.debug("Connected to modem in async_connect")

    def pause_writing(self):
        """Pause writing to the transport."""
        asyncio.ensure_future(self._stop_writer())

    def resume_writing(self):
        """Resume writing to the transport."""
        self._start_writer()

    def close(self):
        """Close the serial transport."""
        self._should_reconnect = False
        asyncio.ensure_future(self._stop_writer())
        if self._transport:
            self._transport.close()

    def _start_writer(self, *args, **kwargs):
        """Start the message writer."""
        if self._transport and not self._transport.is_closing():
            _LOGGER.debug("Scheduling the writer")
            while not self._message_queue.empty():
                self._message_queue.get_nowait()
            self._writer_task = asyncio.create_task(self._write_messages())
            self._writer_task.add_done_callback(self._start_writer)
        else:
            _LOGGER.debug("Did not schedule the writer since we are closing")

    async def _stop_writer(self):
        """Stop the writer task."""
        if self._writer_task:
            self._writer_task.remove_done_callback(self._start_writer)
        await self._message_queue.put((0, None))

    def write(self, msg, priority=5):
        """Prepare data for writing to the transport.

        Data is actually written by _write_message to ensure a pause between writes.
        This approach minimizes NAK messages. This also allows for some messages
        to be lower priority such as 'Load ALDB' versus higher priority such as
        'Set Light Level'.
        """
        self._message_queue.put_nowait((priority, msg))

    async def _write_messages(self):
        """Write data to the transport."""
        await asyncio.sleep(0.1)
        if self._writer_lock.locked():
            _LOGGER.debug("Writer still running")
            return

        async with self._writer_lock:
            _LOGGER.debug("Modem writer started.")
            try:
                while self._transport and not self._transport.is_closing():
                    _, msg = await self._message_queue.get()
                    if msg is None:
                        return
                    _LOGGER_MSG.debug("TX: %s", repr(msg))
                    if (
                        _LOGGER_MSG.level == 0 or _LOGGER_MSG.level > logging.DEBUG
                    ) and (
                        _LOGGER_PYINSTEON.level == 0
                        or _LOGGER_PYINSTEON.level > logging.DEBUG
                    ):
                        for addr in _get_addresses_in_msg(msg):
                            logger = logging.getLogger(f"pyinsteon.{Address(addr).id}")
                            logger.debug("TX: %s", repr(msg))
                    while not self._last_message.empty():
                        self._last_message.get()
                    self._last_message.put(msg)
                    await self._transport.async_write(msg)
                    await asyncio.sleep(self._transport.write_wait)
            except RuntimeError as error:
                _LOGGER.warning(
                    "Modem writer stopped due to a runtime error: %s", str(error)
                )
        _LOGGER.debug("Modem writer stopped.")



"""KeypadLink toggle button property."""
from . import calc_toggle_mode, get_usable_value
from ..constants import ToggleMode
from ..utils import bit_is_set, set_bit
from .derived_property import DerivedProperty


class ToggleButtonProperty(DerivedProperty):
    """KeypadLink toggle button property."""

    def __init__(
        self, address, name, button, non_toggle_mask_prop, non_toggle_on_off_mask_prop
    ):
        """Init the ToggleButtonProperty class."""
        super().__init__(address, name, ToggleMode, False, False)
        self._button_bit = button - 1
        self._non_toggle_mask_prop = non_toggle_mask_prop
        self._non_toggle_on_off_mask_prop = non_toggle_on_off_mask_prop

    @property
    def value(self):
        """Return the toggle mode of the undlying properties."""
        if not self.is_loaded:
            return None
        toggle_value = bit_is_set(self._non_toggle_mask_prop.value, self._button_bit)
        toggle_on_value = bit_is_set(
            self._non_toggle_on_off_mask_prop.value, self._button_bit
        )
        return calc_toggle_mode(toggle_value, toggle_on_value)

    @property
    def new_value(self):
        """Return the modified toggle mode of the button."""
        if not self.is_dirty:
            return None

        toggle_prop_value = get_usable_value(self._non_toggle_mask_prop)
        toggle_on_off_value = get_usable_value(self._non_toggle_on_off_mask_prop)
        toggle = bit_is_set(toggle_prop_value, self._button_bit)
        toggle_on = bit_is_set(toggle_on_off_value, self._button_bit)
        return calc_toggle_mode(toggle, toggle_on)

    @new_value.setter
    def new_value(self, value):
        """Set the new toggle mode value of the button."""
        if not self.is_loaded:
            return

        if value is None:
            self._non_toggle_mask_prop.new_value = None
            self._non_toggle_on_off_mask_prop.new_value = None
            return

        toggle_mode = ToggleMode(value)
        if toggle_mode == self.value:
            self._non_toggle_mask_prop.new_value = None
            self._non_toggle_on_off_mask_prop.new_value = None
            return

        non_toggle_mask = self._non_toggle_mask_prop
        on_only_mask = self._non_toggle_on_off_mask_prop
        non_toggle_val = get_usable_value(non_toggle_mask)
        on_only_val = get_usable_value(on_only_mask)
        if toggle_mode == ToggleMode.TOGGLE:
            non_toggle_mask.new_value = set_bit(non_toggle_val, self._button_bit, False)
            on_only_mask.new_value = set_bit(on_only_val, self._button_bit, False)
        elif toggle_mode == ToggleMode.ON_ONLY:
            non_toggle_mask.new_value = set_bit(non_toggle_val, self._button_bit, True)
            on_only_mask.new_value = set_bit(on_only_val, self._button_bit, True)
        else:
            non_toggle_mask.new_value = set_bit(non_toggle_val, self._button_bit, True)
            on_only_mask.new_value = set_bit(on_only_val, self._button_bit, False)

    @property
    def is_dirty(self):
        """Return the change state of the underlying properties."""
        return (
            self._non_toggle_mask_prop.is_dirty
            or self._non_toggle_on_off_mask_prop.is_dirty
        )

    @property
    def is_loaded(self):
        """Return the load status of the underlying properties."""
        return (
            self._non_toggle_mask_prop.is_loaded
            and self._non_toggle_on_off_mask_prop.is_loaded
        )



"""Create a topic and a direct message."""
import logging
from math import ceil

from .. import pub
from ..address import Address
from ..commands import commands
from ..constants import (
    DeviceCategory,
    DoorCommand,
    IOModuleControlCommand,
    PoolControlCommand,
    RampRate,
    SprinklerControlCommand,
    ThermostatCommand,
    ThermostatZoneInfo,
    WindowCoveringCommand,
)
from ..data_types.all_link_record_flags import AllLinkRecordFlags
from ..data_types.io_sensor_config_flags import IOPortConfigFlags
from ..data_types.message_flags import MessageFlags
from ..data_types.user_data import UserData
from ..topics import (
    ASSIGN_TO_ALL_LINK_GROUP,
    BEEP,
    BRIGHTEN_ONE_STEP,
    DELETE_FROM_ALL_LINK_GROUP,
    DEVICE_TEXT_STRING_REQUEST,
    DIM_ONE_STEP,
    DOOR_CONTROL,
    DOOR_STATUS_REPORT,
    ENTER_LINKING_MODE,
    ENTER_UNLINKING_MODE,
    EXTENDED_GET_SET,
    EXTENDED_READ_WRITE_ALDB,
    EXTENDED_TRIGGER_ALL_LINK,
    FACTORY_RESET,
    FX_USERNAME,
    GET_INSTEON_ENGINE_VERSION,
    GET_OPERATING_FLAGS,
    GROUP_OFF,
    GROUP_ON,
    ID_REQUEST,
    INSTANT_CHANGE,
    IO_ALARM_DATA_REQUEST,
    IO_GET_SENSOR_ALARM_DELTA,
    IO_GET_SENSOR_VALUE,
    IO_MODULE_CONTROL,
    IO_OUTPUT_OFF,
    IO_OUTPUT_ON,
    IO_READ_CONFIGURATION_PORT,
    IO_READ_INPUT_PORT,
    IO_SET_SENSOR_1_NOMINAL_VALUE,
    IO_SET_SENSOR_NOMINAL_VALUE,
    IO_WRITE_CONFIGURATION_PORT,
    IO_WRITE_OUTPUT_PORT,
    NIGHT_MODE_OFF,
    NIGHT_MODE_ON,
    OFF,
    OFF_AT_RAMP_RATE,
    OFF_FAST,
    ON,
    ON_AT_RAMP_RATE,
    ON_FAST,
    PEEK,
    PEEK_INTERNAL,
    PING,
    POKE,
    POKE_INTERNAL,
    POOL_CONTROL,
    POOL_DEVICE_OFF,
    POOL_DEVICE_ON,
    POOL_SET_DEVICE_HYSTERESIS,
    POOL_SET_DEVICE_TEMPERATURE,
    POOL_TEMPERATURE_DOWN,
    POOL_TEMPERATURE_UP,
    PRODUCT_DATA_REQUEST,
    SET_ADDRESS_MSB,
    SET_ALL_LINK,
    SET_ALL_LINK_COMMAND_ALIAS,
    SET_DEVICE_TEXT_STRING,
    SET_OPERATING_FLAGS,
    SET_SPRINKLER_PROGRAM,
    SET_STATUS,
    SPRINKLER_CONTROL,
    SPRINKLER_GET_PROGRAM_REQUEST,
    SPRINKLER_PROGRAM_OFF,
    SPRINKLER_PROGRAM_ON,
    SPRINKLER_VALVE_OFF,
    SPRINKLER_VALVE_ON,
    STATUS_REQUEST,
    THERMOSTAT_CONTROL,
    THERMOSTAT_GET_ZONE_INFORMATION,
    THERMOSTAT_SET_COOL_SETPOINT,
    THERMOSTAT_SET_HEAT_SETPOINT,
    THERMOSTAT_TEMPERATURE_DOWN,
    THERMOSTAT_TEMPERATURE_UP,
    WINDOW_COVERING_CONTROL,
    WINDOW_COVERING_POSITION,
)
from ..utils import subscribe_topic
from .messages.outbound import send_extended, send_standard
from .topic_converters import topic_to_command_handler, topic_to_message_type

_LOGGER = logging.getLogger(__name__)

COMMAND_REGISTER = {}


def register_command_handlers():
    """Register outbound handlers."""
    for topic, func in COMMAND_REGISTER.items():
        subscribe_topic(func, topic)


# The following messages are all send_standard or send_extended messages
# The topis is based on the cmd1, cmd2 and extended message flags values


def _create_direct_message(
    topic, address, cmd2=None, user_data=None, crc=False, priority=5
):
    main_topic = topic.name.split(".")[1]
    command = commands.get_command(main_topic)
    extended = user_data is not None
    cmd2 = command.cmd2 if command.cmd2 is not None else cmd2 if cmd2 is not None else 0
    msg_type = topic_to_message_type(topic)
    flags = MessageFlags.create(msg_type, extended, 3, 3)
    if extended:
        if crc:
            user_data.set_crc(command.cmd1, cmd2)
        else:
            user_data.set_checksum(command.cmd1, cmd2)
        send_extended(
            address=address,
            cmd1=command.cmd1,
            cmd2=cmd2,
            flags=flags,
            user_data=user_data,
            topic=topic,
            priority=priority,
        )
    else:
        send_standard(
            address=address,
            cmd1=command.cmd1,
            cmd2=cmd2,
            flags=flags,
            topic=topic,
            priority=priority,
        )


@topic_to_command_handler(
    register_list=COMMAND_REGISTER, topic=ASSIGN_TO_ALL_LINK_GROUP
)
def assign_to_all_link_group(address: Address, group: int, topic=pub.AUTO_TOPIC):
    """Create a ASSIGN_TO_ALL_LINK_GROUP command."""
    _create_direct_message(topic=topic, address=address, cmd2=group)


@topic_to_command_handler(
    register_list=COMMAND_REGISTER, topic=DELETE_FROM_ALL_LINK_GROUP
)
def delete_from_all_link_group(address: Address, group: int, topic=pub.AUTO_TOPIC):
    """Create a DELETE_FROM_ALL_LINK_GROUP command."""
    _create_direct_message(topic=topic, address=address, cmd2=group)


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=PRODUCT_DATA_REQUEST)
def product_data_request(address: Address, topic=pub.AUTO_TOPIC):
    """Create a PRODUCT_DATA_REQUEST command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=FX_USERNAME)
def fx_username(address: Address, topic=pub.AUTO_TOPIC):
    """Create a FX_USERNAME command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(
    register_list=COMMAND_REGISTER, topic=DEVICE_TEXT_STRING_REQUEST
)
def device_text_string_request(address: Address, topic=pub.AUTO_TOPIC):
    """Create a DEVICE_TEXT_STRING_REQUEST command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=SET_DEVICE_TEXT_STRING)
def set_device_text_string(address: Address, value: str, topic=pub.AUTO_TOPIC):
    """Create a SET_DEVICE_TEXT_STRING command."""
    user_data = UserData()
    for item in range(0, min(len(value), 14)):
        user_data[f"d{item + 1}"] = ord(value[item])
    _create_direct_message(topic=topic, address=address, user_data=user_data)


@topic_to_command_handler(
    register_list=COMMAND_REGISTER, topic=SET_ALL_LINK_COMMAND_ALIAS
)
def set_all_link_command_alias(
    address: Address, cmd: int, alias: int, extended: bool, topic=pub.AUTO_TOPIC
):
    """Create a SET_ALL_LINK_COMMAND_ALIAS command."""
    alias_hi = alias >> 8
    alias_lo = alias & 0xFF
    data4 = 1 if extended else 0
    user_data = UserData({"d1": cmd, "d2": alias_hi, "d3": alias_lo, "d4": data4})
    _create_direct_message(topic=topic, address=address, user_data=user_data)


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=SET_ALL_LINK)
def set_all_link(address: Address, topic=pub.AUTO_TOPIC):
    """Create a SET_ALL_LINK command."""
    user_data = UserData()
    _create_direct_message(topic=topic, address=address, user_data=user_data)


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=ENTER_LINKING_MODE)
def enter_linking_mode(
    address: Address, group: int, extended: bool, topic=pub.AUTO_TOPIC
):
    """Create a ENTER_LINKING_MODE command."""
    if extended:
        user_data = UserData()
    else:
        user_data = None
    _create_direct_message(
        topic=topic, address=address, cmd2=group, user_data=user_data
    )


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=ENTER_UNLINKING_MODE)
def enter_unlinking_mode(
    address: Address, group: int, extended: bool, topic=pub.AUTO_TOPIC
):
    """Create a ENTER_UNLINKING_MODE command."""
    if extended:
        user_data = UserData()
    else:
        user_data = None
    _create_direct_message(
        topic=topic, address=address, cmd2=group, user_data=user_data
    )


@topic_to_command_handler(
    register_list=COMMAND_REGISTER, topic=GET_INSTEON_ENGINE_VERSION
)
def get_insteon_engine_version(address: Address, topic=pub.AUTO_TOPIC):
    """Create a GET_INSTEON_ENGINE_VERSION command."""
    _create_direct_message(topic=topic, address=address, cmd2=0x00)


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=PING)
def ping(address: Address, topic=pub.AUTO_TOPIC):
    """Create a PING command."""
    _create_direct_message(topic=topic, address=address, cmd2=0x00)


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=ID_REQUEST)
def id_request(address: Address, topic=pub.AUTO_TOPIC):
    """Create a ID_REQUEST command."""
    _create_direct_message(topic=topic, address=address, cmd2=0x00)


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=ON)
def on(address: Address, on_level: int, group=0, topic=pub.AUTO_TOPIC):
    """Create a ON command."""
    user_data = None
    if group and group > 1:
        user_data = UserData({"d1": group})
    _create_direct_message(
        topic=topic, address=address, cmd2=on_level, user_data=user_data, priority=3
    )


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=ON_FAST)
def on_fast(address: Address, on_level: int, group: int, topic=pub.AUTO_TOPIC):
    """Create a ON_FAST command."""
    user_data = None
    if group and group > 1:
        user_data = UserData({"d1": group})
    _create_direct_message(
        topic=topic, address=address, cmd2=on_level, user_data=user_data, priority=3
    )


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=OFF)
def off(address: Address, group: int, topic=pub.AUTO_TOPIC):
    """Create a OFF command."""
    user_data = None
    if group and group > 1:
        user_data = UserData({"d1": group})
    _create_direct_message(
        topic=topic, address=address, cmd2=0, user_data=user_data, priority=3
    )


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=OFF_FAST)
def off_fast(address: Address, group: int, topic=pub.AUTO_TOPIC):
    """Create a OFF_FAST command."""
    user_data = None
    if group and group > 1:
        user_data = UserData({"d1": group})
    _create_direct_message(
        topic=topic, address=address, cmd2=0, user_data=user_data, priority=3
    )


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=BRIGHTEN_ONE_STEP)
def brighten_one_step(address: Address, topic=pub.AUTO_TOPIC):
    """Create a BRIGHTEN_ONE_STEP command."""
    _create_direct_message(topic=topic, address=address, cmd2=0, priority=3)


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=DIM_ONE_STEP)
def dim_one_step(address: Address, topic=pub.AUTO_TOPIC):
    """Create a DIM_ONE_STEP command."""
    _create_direct_message(topic=topic, address=address, cmd2=0, priority=3)


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=STATUS_REQUEST)
def status_request(address: Address, status_type: int = 0, topic=pub.AUTO_TOPIC):
    """Create a STATUS_REQUEST command."""
    _create_direct_message(topic=topic, address=address, cmd2=status_type, priority=7)


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=GET_OPERATING_FLAGS)
def get_operating_flags(
    address: Address, flags_requested: int, extended: bool, topic=pub.AUTO_TOPIC
):
    """Create a GET_OPERATING_FLAGS command."""
    user_data = UserData() if extended else None
    _create_direct_message(
        topic=topic,
        address=address,
        cmd2=flags_requested,
        priority=7,
        user_data=user_data,
    )


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=SET_OPERATING_FLAGS)
def set_operating_flags(
    address: Address, cmd: int, extended=False, topic=pub.AUTO_TOPIC
):
    """Create a SET_OPERATING_FLAGS command."""
    user_data = UserData() if extended else None
    _create_direct_message(topic=topic, address=address, cmd2=cmd, user_data=user_data)


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=INSTANT_CHANGE)
def instant_change(address: Address, on_level: int, topic=pub.AUTO_TOPIC):
    """Create a INSTANT_CHANGE command."""
    _create_direct_message(topic=topic, address=address, cmd2=on_level, priority=3)


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=SET_STATUS)
def set_status(address: Address, on_level: int, topic=pub.AUTO_TOPIC):
    """Create a SET_STATUS command."""
    _create_direct_message(topic=topic, address=address, cmd2=on_level)


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=SET_ADDRESS_MSB)
def set_address_msb(address: Address, high_byte: int, topic=pub.AUTO_TOPIC):
    """Create a SET_ADDRESS_MSB command."""
    _create_direct_message(topic=topic, address=address, cmd2=high_byte)


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=POKE)
def poke(address: Address, value: int, topic=pub.AUTO_TOPIC):
    """Create a POKE command."""
    _create_direct_message(topic=topic, address=address, cmd2=value)


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=PEEK)
def peek(address: Address, lsb: int, topic=pub.AUTO_TOPIC):
    """Create a PEEK command."""
    _create_direct_message(topic=topic, address=address, cmd2=lsb)


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=PEEK_INTERNAL)
def peek_internal(address: Address, lsb: int, topic=pub.AUTO_TOPIC):
    """Create a PEEK_INTERNAL command."""
    _create_direct_message(topic=topic, address=address, cmd2=lsb)


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=POKE_INTERNAL)
def poke_internal(address: Address, value: int, topic=pub.AUTO_TOPIC):
    """Create a POKE_INTERNAL command."""
    _create_direct_message(topic=topic, address=address, cmd2=value)


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=ON_AT_RAMP_RATE)
def on_at_ramp_rate(
    address: Address, on_level: int, ramp_rate: RampRate, topic=pub.AUTO_TOPIC
):
    """Create a ON_AT_RAMP_RATE command."""
    on_level = max(0x10, on_level & 0xF0)
    ramp_rate = ceil(int(ramp_rate) / 2) + 1 & 0x0F
    cmd2 = on_level + ramp_rate
    _create_direct_message(topic=topic, address=address, cmd2=cmd2, priority=3)


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=EXTENDED_GET_SET)
def extended_get_set(
    address: Address,
    cmd2=0,
    data1=0,
    data2=0,
    data3=0,
    data4=0,
    data5=0,
    data6=0,
    data7=0,
    data8=0,
    data9=0,
    data10=0,
    data11=0,
    data12=0,
    data13=0,
    data14=0,
    crc=False,
    topic=pub.AUTO_TOPIC,
):
    """Create a EXTENDED_GET_SET command."""
    data = {}
    items = locals()
    for index in range(1, 15):
        item = items[f"data{index}"]
        data[f"d{index}"] = item if item is not None else 0
    user_data = UserData(data)
    _create_direct_message(
        topic=topic, address=address, cmd2=cmd2, user_data=user_data, crc=crc
    )


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=OFF_AT_RAMP_RATE)
def off_at_ramp_rate(address: Address, ramp_rate: RampRate, topic=pub.AUTO_TOPIC):
    """Create a OFF_AT_RAMP_RATE command."""
    ramp_rate = ceil(int(ramp_rate) / 2) + 1 & 0x0F
    cmd2 = ramp_rate
    _create_direct_message(topic=topic, address=address, cmd2=cmd2, priority=3)


def _read_aldb(address, mem_addr, num_recs, topic):
    # num_recs = 0 if mem_addr == 0x0000 else 1
    mem_hi = mem_addr >> 8
    mem_lo = mem_addr & 0xFF
    user_data = UserData(
        {"d1": 0x00, "d2": 0x00, "d3": mem_hi, "d4": mem_lo, "d5": num_recs}
    )
    _create_direct_message(topic=topic, address=address, cmd2=0, user_data=user_data)


def _write_aldb(
    address,
    mem_addr,
    controller,
    group,
    target,
    data1,
    data2,
    data3,
    in_use,
    high_water_mark,
    bit5,
    bit4,
    topic,
):
    address = Address(address)
    target = Address(target)
    mem_hi = mem_addr >> 8
    mem_lo = mem_addr & 0xFF
    flags = AllLinkRecordFlags.create(
        in_use=in_use, controller=controller, hwm=high_water_mark, bit5=bit5, bit4=bit4
    )
    user_data = UserData(
        {
            "d2": 0x02,
            "d3": mem_hi,
            "d4": mem_lo,
            "d5": 0x08,
            "d6": int(flags),
            "d7": group,
            "d8": target.high,
            "d9": target.middle,
            "d10": target.low,
            "d11": data1,
            "d12": data2,
            "d13": data3,
        }
    )
    _create_direct_message(topic=topic, address=address, cmd2=0, user_data=user_data)


@topic_to_command_handler(
    register_list=COMMAND_REGISTER, topic=EXTENDED_READ_WRITE_ALDB
)
def extended_read_write_aldb(
    address: Address,
    action: int,
    mem_addr: int,
    num_recs: int = 0,
    controller: bool = True,
    group: int = 0x01,
    target: Address = None,
    data1: int = 0x00,
    data2: int = 0x00,
    data3: int = 0x00,
    in_use: bool = True,
    high_water_mark: bool = False,
    bit5: int = 0,
    bit4: int = 0,
    topic=pub.AUTO_TOPIC,
):
    """Create a EXTENDED_READ_WRITE_ALDB command."""
    if action == 0x00:
        _read_aldb(address=address, mem_addr=mem_addr, num_recs=num_recs, topic=topic)
    elif action == 0x02:
        _write_aldb(
            address=address,
            mem_addr=mem_addr,
            controller=controller,
            group=group,
            target=target,
            data1=data1,
            data2=data2,
            data3=data3,
            in_use=in_use,
            high_water_mark=high_water_mark,
            bit5=bit5,
            bit4=bit4,
            topic=topic,
        )


@topic_to_command_handler(
    register_list=COMMAND_REGISTER, topic=EXTENDED_TRIGGER_ALL_LINK
)
def extended_trigger_all_link(
    address: Address, group: int, on_level: int, fast: bool, topic=pub.AUTO_TOPIC
):
    """Create a EXTENDED_TRIGGER_ALL_LINK command."""
    use_on_level = on_level not in [0x00, 0xFF]
    user_data = UserData(
        {
            "d1": group,
            "d2": 1 if use_on_level else 0,
            "d3": on_level if use_on_level else 0,
            "d4": 0x13 if on_level == 0 else 0x11,
            "d5": 0,
            "d6": 1 if fast else 0,
        }
    )
    _create_direct_message(topic=topic, address=address, cmd2=0, user_data=user_data)


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=BEEP)
def beep(address: Address, topic=pub.AUTO_TOPIC):
    """Create a BEEP command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=GROUP_ON)
def group_on(address: Address, group: int, topic=pub.AUTO_TOPIC):
    """Create a GROUP ON command."""
    _create_direct_message(topic=topic, address=address, cmd2=group)


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=GROUP_OFF)
def group_off(address: Address, group: int, topic=pub.AUTO_TOPIC):
    """Create a GROUP OFF command."""
    _create_direct_message(topic=topic, address=address, cmd2=group)


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=FACTORY_RESET)
def factory_reset(
    address: Address, cat: DeviceCategory, subcat: int, topic=pub.AUTO_TOPIC
):
    """Create a FACTORY RESET command."""
    user_data = UserData(
        {
            "d1": int(cat),
            "d2": subcat,
        }
    )
    _create_direct_message(topic=topic, address=address, user_data=user_data)


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=NIGHT_MODE_OFF)
def night_mode_off(address: Address, topic=pub.AUTO_TOPIC):
    """Create a NIGHT MODE OFF command."""
    _create_direct_message(topic=topic, address=address, cmd2=0x00)


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=NIGHT_MODE_ON)
def night_mode_on(address: Address, topic=pub.AUTO_TOPIC):
    """Create a BEEP command."""
    _create_direct_message(topic=topic, address=address, cmd2=0x00)


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=SET_SPRINKLER_PROGRAM)
def set_sprinkler_program(
    address: Address, program: int, timers: iter, topic=pub.AUTO_TOPIC
):
    """Create a SET_SPRINKLER_PROGRAM command."""
    user_data = UserData()
    curr_timer = 1
    for timer in timers:
        user_data[f"d{curr_timer}"] = timer
        curr_timer += 1
    _create_direct_message(
        topic=topic, address=address, cmd2=program, user_data=user_data
    )


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=SPRINKLER_VALVE_ON)
def sprinkler_valve_on(address: Address, valve: int, topic=pub.AUTO_TOPIC):
    """Create a SPRINKLER_VALVE_ON command."""
    _create_direct_message(topic=topic, address=address, cmd2=valve)


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=SPRINKLER_VALVE_OFF)
def sprinkler_valve_off(address: Address, valve: int, topic=pub.AUTO_TOPIC):
    """Create a SPRINKLER_VALVE_OFF command."""
    _create_direct_message(topic=topic, address=address, cmd2=valve)


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=SPRINKLER_PROGRAM_ON)
def sprinkler_program_on(address: Address, program: int, topic=pub.AUTO_TOPIC):
    """Create a SPRINKLER_PROGRAM_ON command."""
    _create_direct_message(topic=topic, address=address, cmd2=program)


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=SPRINKLER_PROGRAM_OFF)
def sprinkler_program_off(address: Address, program: int, topic=pub.AUTO_TOPIC):
    """Create a SPRINKLER_PROGRAM_OFF command."""
    _create_direct_message(topic=topic, address=address, cmd2=program)


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=SPRINKLER_CONTROL)
def sprinkler_control(
    address: Address, command: SprinklerControlCommand, topic=pub.AUTO_TOPIC
):
    """Create a SPRINKLER_CONTROL command."""
    _create_direct_message(topic=topic, address=address, cmd2=int(command))


@topic_to_command_handler(
    register_list=COMMAND_REGISTER, topic=SPRINKLER_GET_PROGRAM_REQUEST
)
def sprinkler_get_program_request(address: Address, program: int, topic=pub.AUTO_TOPIC):
    """Create a SPRINKLER_GET_PROGRAM_REQUEST command."""
    _create_direct_message(topic=topic, address=address, cmd2=program)


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=IO_OUTPUT_ON)
def io_output_on(address: Address, output: int, topic=pub.AUTO_TOPIC):
    """Create a IO_OUTPUT_ON command."""
    _create_direct_message(topic=topic, address=address, cmd2=output)


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=IO_OUTPUT_OFF)
def io_output_off(address: Address, output: int, topic=pub.AUTO_TOPIC):
    """Create a IO_OUTPUT_OFF command."""
    _create_direct_message(topic=topic, address=address, cmd2=output)


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=IO_ALARM_DATA_REQUEST)
def io_alarm_data_request(address: Address, topic=pub.AUTO_TOPIC):
    """Create a IO_ALARM_DATA_REQUEST command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=IO_WRITE_OUTPUT_PORT)
def io_write_output_port(address: Address, value: int, topic=pub.AUTO_TOPIC):
    """Create a IO_WRITE_OUTPUT_PORT command."""
    _create_direct_message(topic=topic, address=address, cmd2=value)


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=IO_READ_INPUT_PORT)
def io_read_input_port(address: Address, topic=pub.AUTO_TOPIC):
    """Create a IO_READ_INPUT_PORT command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=IO_GET_SENSOR_VALUE)
def io_get_sensor_value(address: Address, sensor: int, topic=pub.AUTO_TOPIC):
    """Create a IO_GET_SENSOR_VALUE command."""
    _create_direct_message(topic=topic, address=address, cmd2=sensor)


@topic_to_command_handler(
    register_list=COMMAND_REGISTER, topic=IO_SET_SENSOR_1_NOMINAL_VALUE
)
def io_set_sensor_1_nominal_value(address: Address, value: int, topic=pub.AUTO_TOPIC):
    """Create a IO_SET_SENSOR_1_NOMINAL_VALUE command."""
    _create_direct_message(topic=topic, address=address, cmd2=value)


@topic_to_command_handler(
    register_list=COMMAND_REGISTER, topic=IO_SET_SENSOR_NOMINAL_VALUE
)
def io_set_sensor_nominal_value(
    address: Address, sensor: int, value: int, topic=pub.AUTO_TOPIC
):
    """Create a IO_SET_SENSOR_NOMINAL_VALUE command."""
    user_data = UserData({"d1": value})
    _create_direct_message(
        topic=topic, address=address, cmd2=sensor, user_data=user_data
    )


@topic_to_command_handler(
    register_list=COMMAND_REGISTER, topic=IO_GET_SENSOR_ALARM_DELTA
)
def io_get_sensor_alarm_delta(
    address: Address, sensor: int, delta: int, topic=pub.AUTO_TOPIC
):
    """Create a IO_GET_SENSOR_ALARM_DELTA command."""
    negative = delta < 0
    sensor = sensor & 0x0F
    delta = (abs(delta) << 4) & 0x70
    direction = 0x80 if negative else 0
    cmd2 = sensor + delta + direction
    _create_direct_message(topic=topic, address=address, cmd2=cmd2)


@topic_to_command_handler(
    register_list=COMMAND_REGISTER, topic=IO_WRITE_CONFIGURATION_PORT
)
def io_write_configuration_port(
    address: Address,
    port_flags: IOPortConfigFlags,
    topic=pub.AUTO_TOPIC,
):
    """Create a IO_WRITE_CONFIGURATION_PORT command."""
    cmd2 = int(port_flags)
    _create_direct_message(topic=topic, address=address, cmd2=cmd2)


@topic_to_command_handler(
    register_list=COMMAND_REGISTER, topic=IO_READ_CONFIGURATION_PORT
)
def io_read_configuration_port(address: Address, topic=pub.AUTO_TOPIC):
    """Create a IO_READ_CONFIGURATION_PORT command."""
    _create_direct_message(topic=topic, address=address)


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=IO_MODULE_CONTROL)
def io_module_control(
    address: Address, command: IOModuleControlCommand, topic=pub.AUTO_TOPIC
):
    """Create a IO_MODULE_CONTROL command."""
    _create_direct_message(topic=topic, address=address, cmd2=int(command))


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=POOL_DEVICE_ON)
def pool_device_on(address: Address, device_num: int, topic=pub.AUTO_TOPIC):
    """Create a POOL_DEVICE_ON command."""
    _create_direct_message(topic=topic, address=address, cmd2=device_num)


@topic_to_command_handler(
    register_list=COMMAND_REGISTER, topic=POOL_SET_DEVICE_TEMPERATURE
)
def pool_set_device_temperature(
    address: Address, device_num: int, degrees: int, topic=pub.AUTO_TOPIC
):
    """Create a POOL_SET_DEVICE_TEMPERATURE command."""
    user_data = UserData({"d1": device_num, "d2": degrees})
    _create_direct_message(topic=topic, address=address, user_data=user_data)


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=POOL_DEVICE_OFF)
def pool_device_off(address: Address, device_num: int, topic=pub.AUTO_TOPIC):
    """Create a POOL_DEVICE_OFF command."""
    _create_direct_message(topic=topic, address=address, cmd2=device_num)


@topic_to_command_handler(
    register_list=COMMAND_REGISTER, topic=POOL_SET_DEVICE_HYSTERESIS
)
def pool_set_device_hysteresis(
    address: Address, device_num: int, hysteresis: int, topic=pub.AUTO_TOPIC
):
    """Create a POOL_SET_DEVICE_HYSTERESIS command."""
    user_data = UserData({"d1": device_num, "d2": hysteresis})
    _create_direct_message(topic=topic, address=address, user_data=user_data)


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=POOL_TEMPERATURE_UP)
def pool_temperature_up(address: Address, degrees: int, topic=pub.AUTO_TOPIC):
    """Create a POOL_TEMPERATURE_UP command."""
    cmd2 = degrees * 2
    _create_direct_message(topic=topic, address=address, cmd2=cmd2)


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=POOL_TEMPERATURE_DOWN)
def pool_temperature_down(address: Address, degrees: int, topic=pub.AUTO_TOPIC):
    """Create a POOL_TEMPERATURE_DOWN command."""
    cmd2 = degrees * 2
    _create_direct_message(topic=topic, address=address, cmd2=cmd2)


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=POOL_CONTROL)
def pool_control(address: Address, command: PoolControlCommand, topic=pub.AUTO_TOPIC):
    """Create a POOL_CONTROL command."""
    _create_direct_message(topic=topic, address=address, cmd2=int(command))


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=DOOR_CONTROL)
def door_control(address: Address, command: DoorCommand, topic=pub.AUTO_TOPIC):
    """Create a DOOR_CONTROL command."""
    _create_direct_message(topic=topic, address=address, cmd2=int(command))


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=DOOR_STATUS_REPORT)
def door_status_report(address: Address, command: DoorCommand, topic=pub.AUTO_TOPIC):
    """Create a DOOR_STATUS_REPORT command."""
    _create_direct_message(topic=topic, address=address, cmd2=int(command))


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=WINDOW_COVERING_CONTROL)
def window_covering_command(
    address: Address, command: WindowCoveringCommand, topic=pub.AUTO_TOPIC
):
    """Create a WINDOW_COVERING_CONTROL command."""
    _create_direct_message(topic=topic, address=address, cmd2=int(command))


@topic_to_command_handler(
    register_list=COMMAND_REGISTER, topic=WINDOW_COVERING_POSITION
)
def window_covering_position(address: Address, position: int, topic=pub.AUTO_TOPIC):
    """Create a WINDOW_COVERING_POSITION command."""
    _create_direct_message(topic=topic, address=address, cmd2=position)


@topic_to_command_handler(
    register_list=COMMAND_REGISTER, topic=THERMOSTAT_TEMPERATURE_UP
)
def thermostat_temperature_up(
    address: Address, degrees: int, zone: int, topic=pub.AUTO_TOPIC
):
    """Create a THERMOSTAT_TEMPERATURE_UP command."""
    degrees_value = int(round(degrees * 2, 0))
    if not zone:
        cmd2 = degrees_value
        user_data = UserData()
    else:
        cmd2 = zone
        user_data = UserData({"d1": degrees_value})
    _create_direct_message(topic=topic, address=address, cmd2=cmd2, user_data=user_data)


@topic_to_command_handler(
    register_list=COMMAND_REGISTER, topic=THERMOSTAT_TEMPERATURE_DOWN
)
def thermostat_temperature_down(
    address: Address, degrees: int, zone: int, topic=pub.AUTO_TOPIC
):
    """Create a THERMOSTAT_TEMPERATURE_DOWN command."""
    degrees_value = int(round(degrees * 2, 0))
    if not zone:
        cmd2 = degrees_value
        user_data = UserData()
    else:
        cmd2 = zone
        user_data = UserData({"d1": degrees_value})
    _create_direct_message(topic=topic, address=address, cmd2=cmd2, user_data=user_data)


@topic_to_command_handler(
    register_list=COMMAND_REGISTER, topic=THERMOSTAT_GET_ZONE_INFORMATION
)
def thermostat_get_zone_information(
    address: Address, zone: int, info: ThermostatZoneInfo, topic=pub.AUTO_TOPIC
):
    """Create a THERMOSTAT_GET_ZONE_INFORMATION command.

    zone: (int) 0 to 31

    info: (int)
        0 = Temperature
        1 = Setpoint
        2 = Deadband
        3 = Humidity
    """
    zone = zone & 0x1F
    info_bits = (info << 5) & 0x60
    cmd2 = info_bits + zone
    _create_direct_message(topic=topic, address=address, cmd2=cmd2)


@topic_to_command_handler(register_list=COMMAND_REGISTER, topic=THERMOSTAT_CONTROL)
def thermostat_control(
    address: Address, thermostat_mode: ThermostatCommand, topic=pub.AUTO_TOPIC
):
    """Create a THERMOSTAT_CONTROL command."""
    user_data = UserData()
    _create_direct_message(
        topic=topic, address=address, cmd2=int(thermostat_mode), user_data=user_data
    )


@topic_to_command_handler(
    register_list=COMMAND_REGISTER, topic=THERMOSTAT_SET_COOL_SETPOINT
)
def thermostat_set_cool_setpoint(
    address: Address, degrees: int, zone: int, deadband: int, topic=pub.AUTO_TOPIC
):
    """Create a THERMOSTAT_SET_COOL_SETPOINT command."""
    if not zone:
        cmd2 = int(degrees * 2)
        user_data = UserData()
    else:
        cmd2 = zone
        user_data = UserData({"d1": degrees * 2, "d2": deadband * 2})
    _create_direct_message(topic=topic, address=address, cmd2=cmd2, user_data=user_data)


@topic_to_command_handler(
    register_list=COMMAND_REGISTER, topic=THERMOSTAT_SET_HEAT_SETPOINT
)
def thermostat_set_heat_setpoint(
    address: Address, degrees: int, zone: int, deadband: int, topic=pub.AUTO_TOPIC
):
    """Create a THERMOSTAT_SET_HEAT_SETPOINT command."""
    if not zone:
        cmd2 = int(degrees * 2)
        user_data = UserData()
    else:
        cmd2 = zone
        user_data = UserData({"d1": degrees * 2, "d2": deadband * 2})
    _create_direct_message(topic=topic, address=address, cmd2=cmd2, user_data=user_data)



"""Manage outbound ON command to a device."""

from ...topics import ON_FAST
from .direct_command import DirectCommandHandlerBase


class OnFastCommand(DirectCommandHandlerBase):
    """Manage an outbound ON command to a device."""

    def __init__(self, address, group):
        """Init the OnFastCommand class."""
        super().__init__(topic=ON_FAST, address=address, group=group)

    # pylint: disable=arguments-differ
    async def async_send(self, on_level=0xFF):
        """Send the ON FAST command async."""
        return await super().async_send(on_level=on_level, group=self._group)

    def _update_subscribers_on_direct_ack(
        self, cmd1, cmd2, target, user_data, hops_left
    ):
        """Update subscribers."""
        self._call_subscribers(on_level=cmd2 if cmd2 else 0xFF)



"""Manage outbound ON command to a device."""

from ...topics import NIGHT_MODE_OFF
from .direct_command import DirectCommandHandlerBase


class NightModeOffCommand(DirectCommandHandlerBase):
    """Manage an outbound NIGHT_MODE_OFF command to a device."""

    def __init__(self, address):
        """Init the NightModeOffCommand class."""
        super().__init__(topic=NIGHT_MODE_OFF, address=address)

    # pylint: disable=arguments-differ
    async def async_send(self):
        """Send the NIGHT_MODE_OFF command async."""
        return await super().async_send()

    def _update_subscribers_on_direct_ack(
        self, cmd1, cmd2, target, user_data, hops_left
    ):
        """Update subscribers."""
        self._call_subscribers(night_mode=False)



"""Manage outbound ON command to a device."""

import asyncio

from .. import ack_handler, status_handler
from ... import pub
from ...constants import MessageFlagType, ResponseStatus
from ...topics import STATUS_REQUEST
from .direct_command import DirectCommandHandlerBase


class StatusRequestCommand(DirectCommandHandlerBase):
    """Manage an outbound Status command to a device."""

    def __init__(self, address, status_type: int = 0):
        """Init the OnLevelCommand class."""
        self._status_type = status_type
        super().__init__(topic=STATUS_REQUEST, address=address, group=None)
        self._subscriber_topic = f"handler.{self._address.id}.{self._status_type}.{STATUS_REQUEST}.{str(MessageFlagType.DIRECT).lower()}"

    @property
    def status_type(self):
        """Return the status type."""
        return self._status_type

    # pylint: disable=arguments-differ, useless-super-delegation
    async def async_send(self):
        """Send the ON command async."""
        return await super().async_send(status_type=self._status_type)

    @ack_handler
    async def async_handle_ack(self, cmd1, cmd2, user_data):
        """Handle the message ACK."""
        if cmd2 == self._status_type:
            await super().async_handle_ack(cmd1=cmd1, cmd2=cmd2, user_data=user_data)

    @status_handler
    async def async_handle_direct_ack(self, topic=pub.AUTO_TOPIC, **kwargs):
        """Handle the Status Request response direct ACK.

        This handler listens to all topics for a device therefore we need to
        confirm the message is a status response.
        """
        # Need to make sure the ACK has time to aquire the lock
        await asyncio.sleep(0.06)
        if not self._response_lock.locked():
            return

        msg_type = topic.name.split(".")[-1]
        if msg_type != str(MessageFlagType.DIRECT_ACK):
            return

        self._direct_response.put_nowait(ResponseStatus.SUCCESS)

        cmd1 = kwargs.get("cmd1")
        cmd2 = kwargs.get("cmd2")
        self._call_subscribers(db_version=cmd1, status=cmd2)



"""Thermostat temperature up command."""
from ...constants import ThermostatMode
from ...topics import THERMOSTAT_CONTROL
from .direct_command import DirectCommandHandlerBase

MODE_MAP = {
    ThermostatMode.HEAT: 0x04,
    ThermostatMode.COOL: 0x05,
    ThermostatMode.AUTO: 0x06,
    ThermostatMode.FAN_ALWAYS_ON: 0x07,
    ThermostatMode.FAN_AUTO: 0x08,
    ThermostatMode.OFF: 0x09,
}

THERMOSTAT_MODE_MAP = {
    0x04: ThermostatMode.HEAT,
    0x05: ThermostatMode.COOL,
    0x06: ThermostatMode.AUTO,
    0x07: ThermostatMode.FAN_ALWAYS_ON,
    0x08: ThermostatMode.FAN_AUTO,
    0x09: ThermostatMode.OFF,
}


class ThermostatModeCommand(DirectCommandHandlerBase):
    """Manage an outbound THERMOSTAT_TEMPERATURE_DOWN command to a device."""

    def __init__(self, address):
        """Init the TemperatureUpCommand class."""
        super().__init__(topic=THERMOSTAT_CONTROL, address=address)

    # pylint: disable=arguments-differ
    async def async_send(self, thermostat_mode):
        """Send the OFF command async."""
        thermostat_mode = ThermostatMode(thermostat_mode)
        send_mode = MODE_MAP.get(thermostat_mode)
        if send_mode is None:
            raise ValueError("Invalid thermostat mode")
        return await super().async_send(thermostat_mode=send_mode)

    def _update_subscribers_on_direct_ack(
        self, cmd1, cmd2, target, user_data, hops_left
    ):
        """Update subscribers."""
        thermostat_mode = THERMOSTAT_MODE_MAP.get(cmd2)
        if thermostat_mode is None:
            return
        self._call_subscribers(thermostat_mode=thermostat_mode)



"""Manage inbound ON command from device."""
from ...topics import OFF_FAST
from .broadcast_command import BroadcastCommandHandlerBase


class OffFastInbound(BroadcastCommandHandlerBase):
    """Off Level command inbound."""

    def __init__(self, address, group):
        """Init the OffFastInbound class."""
        super().__init__(topic=OFF_FAST, address=address, group=group)

    def _handle_message_received(self, cmd1, cmd2, target, user_data, hops_left):
        """Handle the OFF command from a device."""
        self._call_subscribers(on_level=0)



"""Advanced ALDB tools."""
import json
import os

import aiofiles

from .. import devices
from ..aldb.aldb_record import ALDBRecord
from ..constants import ALDBStatus, AllLinkMode, LinkStatus
from ..managers.link_manager import (
    async_cancel_linking_mode,
    async_enter_linking_mode,
    find_broken_links,
)
from ..managers.saved_devices_manager import aldb_rec_to_dict, dict_to_aldb_record
from .aldb import ToolsAldb

LINK_MODE_MAP = {
    "c": AllLinkMode.CONTROLLER,
    "r": AllLinkMode.RESPONDER,
    "e": AllLinkMode.EITHER,
}


class AdvancedTools(ToolsAldb):
    """Advanced ALDB tools."""

    async def do_add_link(
        self,
        address,
        group,
        target,
        link_mode,
        data1=0,
        data2=0,
        data3=0,
        log_stdout=None,
        background=False,
    ):
        """Add a link to a device All-Link Database.

        For modems use the add_im_link command.

        Usage:
            add_device_link [--background | -b] address group target link_mode [data1 data2 data3]

        address: Address of the device to add the link [0 - 255]
        group: Group number of the link [0 - 255]
        target: The device target the link refers to
        link_mode: c or r   Controller or responder link type (c=Controller, r=Responder)
        """

        try:
            addresses = await self._ensure_address(
                address=address,
                name="Address",
                ask_value=not background,
                log_stdout=log_stdout,
                allow_all=False,
                match_device=True,
            )
            if not addresses:
                return
            address = addresses[0]
        except ValueError:
            log_stdout("Invalid device address or device not found")
            return

        device = devices[address]

        if device.aldb.status != ALDBStatus.LOADED:
            log_stdout(
                f"The All-Link Database for device {device.address} must be loaded first."
            )
            return

        record = {
            "link_mode": link_mode,
            "hwm": "n",
            "in_use": "y",
            "group": group,
            "target": target,
            "data1": data1,
            "data2": data2,
            "data3": data3,
        }
        try:
            record = await self._parse_record(log_stdout, not background, **record)
        except ValueError:
            return

        device.aldb.add(
            group=record["group"],
            target=record["target"],
            controller=record["is_controller"],
            data1=record["data1"],
            data2=record["data2"],
            data3=record["data3"],
        )
        success, _ = await device.aldb.async_write()
        if device.is_battery:
            log_stdout(
                f"Device {device.address} is battery operated. The record will be written when the device wakes up."
            )
            return

        if success:
            log_stdout(f"The record was successfully writen to device {device.address}")
            return

        log_stdout(
            f"An issue occured writing to the database of device {device.address}"
        )

    async def do_remove_link(
        self, address, mem_addr, log_stdout=None, background=False
    ):
        """Remove a link from the All-Link Database.

        Usage:
            remove_link [--background | -b] address mem_addr

            address: Address of the device
            mem_addr: Memory address of the link to remove (i.e. 0f7f)
        """
        try:
            addresses = await self._ensure_address(
                address=address,
                name="Address",
                ask_value=not background,
                log_stdout=log_stdout,
                allow_all=False,
                match_device=True,
            )
            if not addresses:
                log_stdout("Address is required")
                return
            address = addresses[0]
        except ValueError:
            log_stdout("Invalid device address or device not found")
            return

        device = devices[address]
        if device.aldb.status != ALDBStatus.LOADED:
            log_stdout(
                f"The All-Link Database for device {device.address} must be loaded first."
            )
            return

        try:
            mem_addr = await self._ensure_hex_byte(
                mem_addr,
                "Memory address",
                not background,
                log_stdout,
                values=device.aldb,
            )
            if mem_addr is None:
                log_stdout("Record memory address is required")
                return

        except (IndexError, ValueError):
            log_stdout("Invalid record memory address")
            return

        device.aldb.remove(mem_addr)
        success, _ = await device.aldb.async_write()
        if device.is_battery:
            log_stdout(
                f"Device {device.address} is battery operated. The record will be removed when the device wakes up."
            )
        elif success:
            log_stdout(
                f"The record was successfully removed from the device {device.address}"
            )
        else:
            log_stdout(
                f"An issue occured writing to the database of device {device.address}"
            )

    def do_find_broken_links(self, log_stdout=None, background=False):
        """Find broken links between devices.

        Useage:
            find_broken_links [--background | -b]
        """
        broken_links = find_broken_links(devices)
        log_stdout("Device   Mem Addr Target    Group Mode Status")
        log_stdout(
            "-------- -------- --------- ----- ---- ----------------------------------------"
        )
        for address, broken_link in broken_links.items():
            for mem_addr, rec, status in broken_link:
                if status == LinkStatus.MISSING_CONTROLLER:
                    status_txt = "Missing controller"
                elif status == LinkStatus.MISSING_RESPONDER:
                    status_txt = "Missing responder"
                elif status == LinkStatus.MISSING_TARGET:
                    status_txt = "Target device not found"
                elif status == LinkStatus.TARGET_DB_NOT_LOADED:
                    status_txt = "Cannot verify - Target ALDB not loaded"
                if rec.is_controller:
                    link_mode = "C"
                else:
                    link_mode = "R"
                log_stdout(
                    f"{str(address)}     {mem_addr:04x} {str(rec.target)} {rec.group:5d}   {link_mode:s} {status_txt:.40s}"
                )

    async def do_change_link(
        self, address, mem_addr, log_stdout=None, background=False, **kwargs
    ):
        """Add a link to a device All-Link Database.

        For modems use the add_im_link command.

        Usage:
            add_device_link [--background | -b] address rec_id field=value [field=value field=value] [force]


        address: Address of the device to add the link [0 - 255]
        rec_id: ALDB record id (i.e. 1fff. Can be found with the print_aldb command)

        Field can be any of the following:
            in_use: Should the record be in use
                y: record is active
                n: record is inactive (same as deleting the record)
            link_mode: Controller or responder link type
                c: Controller
                r: Responder)
            hwm: Is the record the high water mark
                y: record is hwm
                n: record is not hwm
            group: Group number of the link [0 - 255]
            target: The address of the device target the link refers to
            data1: Data field 1 [0 - 255]
            data2: Data field 2 [0 - 255]
            data3: Data feild 3 [0 - 255]

        force: Force the write to happen (overrides ALDB load status)
        """

        try:
            addresses = await self._ensure_address(
                address=address,
                name="Address",
                ask_value=not background,
                log_stdout=log_stdout,
                allow_all=False,
                match_device=True,
            )
            if not addresses:
                return
            address = addresses[0]
        except ValueError:
            log_stdout("Invalid device address or device not found")
            return

        device = devices[address]
        if device.aldb.status != ALDBStatus.LOADED:
            log_stdout(
                f"The All-Link Database for device {device.address} must be loaded first."
            )
            return

        try:
            mem_addr = await self._ensure_hex_byte(
                mem_addr, "Record ID", not background, log_stdout, values=device.aldb
            )
            if mem_addr is None:
                log_stdout("Record ID is required")
                return
        except ValueError:
            log_stdout("Invalid record ID")
            return

        key_values = [
            "in_use",
            "link_mode",
            "hwm",
            "target",
            "group",
            "data1",
            "data2",
            "data3",
            "force",
        ]

        kwargs_found = bool(kwargs)
        changes = {}
        while True:
            if kwargs_found:
                try:
                    arg, value = kwargs.popitem()
                except KeyError:
                    break

            else:
                value = None
                arg = None

            try:
                arg = await self._ensure_string(
                    arg,
                    key_values,
                    "Record value to change (i.e. in_use, data1, etc.)",
                    not background,
                    log_stdout,
                )
            except ValueError:
                log_stdout(f"Invalid record value to change: {arg}")
                return

            if arg is None:
                break

            try:
                if arg in ["in_use", "hwm", "force"]:
                    value = await self._ensure_bool(
                        value, arg, not background, log_stdout=log_stdout
                    )
                if arg == "link_mode":
                    value = await self._ensure_bool(
                        value,
                        arg,
                        not background,
                        log_stdout=log_stdout,
                        true_val="c",
                        values=["c", "r"],
                    )
                if arg == "target":
                    values = await self._ensure_address(
                        value,
                        arg,
                        not background,
                        log_stdout=log_stdout,
                        allow_all=False,
                        match_device=False,
                    )
                    if not values:
                        log_stdout(f"Value is required for {arg}")
                        return
                    value = values[0]

                if arg in ["group", "data1", "data2", "data3"]:
                    value = await self._ensure_byte(
                        value, arg, not background, log_stdout=log_stdout
                    )
            except ValueError:
                log_stdout(f"Invalid value for {arg}")
                return

            if value is None:
                log_stdout(f"Value is required for {arg}")
                return

            changes[arg] = value

        if not changes or ("force" in changes and len(changes) == 1):
            log_stdout("At least one value must be changed")
            return

        device.aldb.modify(
            mem_addr=mem_addr,
            group=changes.get("group"),
            target=changes.get("target"),
            in_use=changes.get("in_use"),
            controller=changes.get("link_mode"),
            high_water_mark=changes.get("hwm"),
            data1=changes.get("data1"),
            data2=changes.get("data2"),
            data3=changes.get("data3"),
        )
        success, _ = await device.aldb.async_write(force=changes.get("force"))
        if device.is_battery:
            log_stdout(
                f"Device {device.address} is battery operated. The record will be written when the device wakes up."
            )
        elif success:
            log_stdout(f"The record was successfully writen to device {device.address}")
        else:
            log_stdout(
                f"An issue occured writing to the database of device {device.address}"
            )
            device.aldb.clear_pending()

    async def do_export_aldb(
        self, address, location=None, filename=None, log_stdout=None, background=False
    ):
        """Export the All-Link Database for a device to a file.

        Usage:
            export_aldb [--background | -b] address [location] [filename]

        address: Address of the device to export
        location: Directory location of where to place the file (default is the current directory).
        filename: Filename of the device. The default file name is `<address>_aldb.json`
        """
        try:
            addresses = await self._ensure_address(
                address=address,
                name="Address",
                ask_value=not background,
                log_stdout=log_stdout,
                allow_all=False,
                match_device=True,
            )
            if not addresses:
                log_stdout("Address is required")
                return
            address = addresses[0]
        except ValueError:
            log_stdout("Invalid device address or device not found")
            return

        device = devices[address]

        if location is None and not background:
            location = await self._get_workdir()
            if location is None:
                log_stdout("File location is required")
                return

        if location is None and self.workdir:
            location = self.workdir

        if location is None:
            log_stdout("File location is required")
            return

        if location == ".":
            location = os.getcwd()

        default_filename = f"{device.address.id}_aldb.json"
        filename = await self._ensure_string(
            filename,
            None,
            "Filename",
            not background,
            self._log_stdout,
            default=default_filename,
        )

        aldb_dict = {}
        for mem in device.aldb:
            rec = device.aldb[mem]
            rec_dict = aldb_rec_to_dict(rec)
            aldb_dict[mem] = rec_dict
        await self._write_aldb_file(log_stdout, aldb_dict, location, filename)

    async def do_replace_aldb(self, address, location, filename):
        """Read and replace the device ALDB with the contents of a file.

        Reads a JSON formated file with All-Link Database records and replaces
        the current records in the device. To create the JSON file use the
        `export_aldb` command.

        WARNING: This method is very dangerous and can make your device non-responsive.

        Usage:
            replace_aldb address [location [filename]]

        address: Address of the device
        location: Directory location of the file (default is the current directory).
        filename: Name of the file to use (default is <address>_aldb.json)
        """

        addresses = await self._ensure_address(
            address=address,
            name="Address",
            ask_value=True,
            log_stdout=self._log_stdout,
            allow_all=False,
            match_device=True,
        )
        if not addresses:
            return
        address = addresses[0]

        device = devices[address]

        if location is None:
            location = await self._get_workdir()
            if location is None:
                self._log_stdout("File location is required")
                return

        if location == ".":
            location = os.getcwd()

        default_filename = f"{device.address.id}_aldb.json"
        filename = await self._ensure_string(
            filename, None, "Filename", True, self._log_stdout, default=default_filename
        )

        try:
            aldb_dict = await self._read_aldb_file(self._log_stdout, location, filename)
        except ValueError:
            self._log_stdout(f"Could not load the file {location}/{filename}")
            return
        aldb_recs = dict_to_aldb_record(aldb_dict)
        aldb_array = [rec for _, rec in aldb_recs.items()]
        self._log_stdout("Preview of changes:")
        self._print_aldb_output(self._log_stdout, device, aldb_array)
        confirm = await self._ensure_bool(
            None, "Do you want to proceed", True, self._log_stdout
        )
        if not confirm:
            return

        for mem_addr in device.aldb:
            device.aldb.remove(mem_addr=mem_addr)

        for rec in aldb_array:
            device.aldb[rec.mem_addr] = rec

        last_rec = aldb_array[-1]
        if not last_rec.is_high_water_mark:
            mem = last_rec.mem_addr - 8
            new_rec = ALDBRecord(mem, False, 0, "00.00.00", 0, 0, 0, False, True, 0, 0)
            device.aldb[mem] = new_rec
        done, not_done = await device.aldb.async_write(force=True)

        if device.is_battery:
            self._log_stdout(
                "This device is battery operated. The ALDB will be written when the device wakes up."
            )
            return

        if not done:
            self._log_stdout("No records were written")
        elif not_done:
            self._log_stdout("Some records not written")
        else:
            self._log_stdout("All records written succesfully")

    async def do_enter_linking_mode(
        self, link_mode, group, log_stdout=None, background=False
    ):
        """Put the modem into linking mode.

        Useage:
            enter_linking_mode [--background | -b] link_mode group

        link_mode: c | r | e  (c = Modem is controller, r = Modem is responder, e = Either)
        group: 0 - 255
        """
        try:
            link_mode = await self._ensure_string(
                value=link_mode,
                values=["c", "r", "e"],
                name="Link mode",
                ask_value=not background,
                log_stdout=log_stdout,
            )
            if link_mode is None:
                log_stdout("Linking mode is required")
                return
        except ValueError:
            log_stdout("Invalid linking mode")
            return

        try:
            group = await self._ensure_byte(group, "Group", not background, log_stdout)
            if group is None:
                log_stdout("Group is required")
                return
        except ValueError:
            log_stdout("Invalid group number")
            return

        await async_enter_linking_mode(
            link_mode=LINK_MODE_MAP[link_mode], group=group, address=None
        )

    async def do_cancel_linking_mode(self):
        """Take the modem out of linking mode.

        Useage:
            cancel_linking_mode
        """
        await async_cancel_linking_mode()

    async def do_find_im_records(
        self, address, group, log_stdout=None, background=False
    ):
        """Find one or more records in the Modem All-Link Database.

        This command does read the Insteon modem rather than read what is currently in memory.

        Useage:
            find_im_records [--background | -b] address group

        address: Insteon address of a device
        group: 0 - 255
        """

        try:
            addresses = await self._ensure_address(
                address=address,
                name="Address",
                ask_value=not background,
                log_stdout=log_stdout,
                allow_all=False,
                match_device=True,
            )
            if not addresses:
                log_stdout("Address is required")
                return
            address = addresses[0]
        except ValueError:
            log_stdout("Invalid device address or device not found")
            return

        try:
            group = await self._ensure_byte(group, "Group", not background, log_stdout)
            if group is None:
                log_stdout("Group number is required")
                return
        except ValueError:
            log_stdout("Invalid group number")
            return

        recs = [
            rec async for rec in devices.modem.aldb.async_find_records(address, group)
        ]
        if not recs:
            log_stdout("No records found")
            return

        self._print_aldb_output(log_stdout, devices.modem, recs)

    async def _write_aldb_file(self, log_stdout, aldb_dict, location, filename):
        device_file = os.path.join(location, filename)
        try:
            async with aiofiles.open(device_file, "w") as afp:
                out_json = json.dumps(aldb_dict, indent=2)
                await afp.write(out_json)
                await afp.flush()
        except FileNotFoundError as ex:
            log_stdout(f"Cannot write to file {device_file}")
            log_stdout(f"Exception: {str(ex)}")

    async def _read_aldb_file(self, log_stdout, location, filename):
        """Load device information from the device info file."""
        aldb_dict = []
        try:
            device_file = os.path.join(location, filename)
            async with aiofiles.open(device_file, "r") as afp:
                json_file = ""
                json_file = await afp.read()
            try:
                aldb_dict = json.loads(json_file)
            except json.decoder.JSONDecodeError:
                log_stdout("Loading ALDB file failed")
        except FileNotFoundError as ex:
            log_stdout("ALDB file not found")
            raise ValueError from ex
        return aldb_dict

    async def _parse_record(self, log_stdout, ask_values, **kwargs):
        """Create a dict of ALDB record values.

        Sets the value to `None` if the value is not received and `ask_values` is `False`.
        Returns `None` if bad values are received and ask_values is `False`.
        """
        link_mode = kwargs.get("link_mode")
        group = kwargs.get("group")
        target = kwargs.get("target")
        data1 = kwargs.get("data1")
        data2 = kwargs.get("data2")
        data3 = kwargs.get("data3")

        name = ""
        try:
            name = "Group"
            group = await self._ensure_byte(group, name, ask_values, log_stdout)
            if group is None:
                log_stdout("Group is required")
                raise ValueError
            name = "Target"

            targets = await self._ensure_address(
                target,
                name,
                ask_values,
                log_stdout,
                allow_all=False,
                match_device=False,
            )
            if not targets:
                log_stdout("Target address is required")
                raise ValueError
            target = targets[0]

            name = "Link mode"
            is_controller = await self._ensure_bool(
                link_mode, name, ask_values, log_stdout, "c", ["r", "c"]
            )
            if is_controller is None:
                log_stdout("Link mode is required")
                raise ValueError

            name = "Data1"
            data1 = await self._ensure_byte(
                data1, name, ask_values, log_stdout, default=0
            )
            name = "Data2"
            data2 = await self._ensure_byte(
                data2, name, ask_values, log_stdout, default=0
            )
            name = "Data3"
            data3 = await self._ensure_byte(
                data3, name, ask_values, log_stdout, default=0
            )

        except ValueError as ex:
            log_stdout(f"Invalid value for {name}")
            raise ValueError from ex

        return {
            "is_in_use": True,
            "is_controller": is_controller,
            "is_high_water_mark": False,
            "group": group,
            "target": target,
            "data1": data1,
            "data2": data2,
            "data3": data3,
        }



"""Base class for the All-Link Database."""

from abc import ABC, abstractmethod
import logging
from typing import List, Tuple

from ..address import Address
from ..constants import ALDBStatus, EngineVersion, ReadWriteMode, ResponseStatus
from ..managers.aldb_write_manager import ALDBWriteException, ALDBWriteManager
from ..topics import ALDB_LINK_CHANGED, ALDB_STATUS_CHANGED, ENGINE_VERSION
from ..utils import publish_topic, subscribe_topic, unsubscribe_topic
from .aldb_record import ALDBRecord, new_aldb_record_from_existing

_LOGGER = logging.getLogger(__name__)
HWM_RECORD = ALDBRecord(
    0x0000,
    controller=False,
    group=0,
    target="000000",
    data1=0,
    data2=0,
    data3=0,
    in_use=False,
    high_water_mark=True,
    bit4=False,
    bit5=False,
)


class ALDBBase(ABC):
    """Represents a base class for a device All-Link database."""

    def __init__(
        self,
        address,
        version=EngineVersion.UNKNOWN,
        mem_addr=0x0FFF,
        write_manager=ALDBWriteManager,
    ):
        """Instantiate the ALL-Link Database object."""
        self._read_write_mode = ReadWriteMode.STANDARD
        self._records = {}
        self._status = ALDBStatus.EMPTY
        self._version = version

        self._address = Address(address)
        self._mem_addr = mem_addr
        self._read_manager = None
        self._write_manager = write_manager(self)
        self._dirty_records = {}
        subscribe_topic(self.update_version, f"{repr(self._address)}.{ENGINE_VERSION}")

    def __len__(self):
        """Return the number of devices in the ALDB."""
        return len(self._records)

    def __iter__(self):
        """Iterate through each ALDB device record."""
        for key in sorted(self._records, reverse=True):
            yield key

    def __getitem__(self, mem_addr):
        """Fetch a device from the ALDB."""
        return self._records.get(mem_addr)

    def __setitem__(self, mem_addr, record):
        """Add or Update a device in the ALDB."""
        if not isinstance(record, ALDBRecord):
            raise ValueError
        if mem_addr == 0:
            mem_addr = self._next_new_mem_addr()
            record.mem_addr = 0
        else:
            record.mem_addr = mem_addr
        self._dirty_records[mem_addr] = record

    def __repr__(self):
        """Human representation of a device from the ALDB."""
        attrs = vars(self)
        return ", ".join(f"{k}: {repr(v)}" for k, v in attrs.items())

    def items(self):
        """Return the memory address an record key value pair."""
        for mem_addr in sorted(self._records, reverse=True):
            yield mem_addr, self._records[mem_addr]

    @property
    def address(self) -> Address:
        """Return the address of the device."""
        return self._address

    @property
    def status(self) -> ALDBStatus:
        """Return the ALDB load status."""
        return self._status

    @property
    def version(self) -> EngineVersion:
        """Return the device engine version."""
        return self._version

    @property
    def first_mem_addr(self):
        """Return the expected memory address of the first record."""
        return self._mem_addr

    @property
    def high_water_mark_mem_addr(self) -> int:
        """Return the High Water Mark record memory address."""
        for mem_addr, rec in self.items():
            if rec.is_high_water_mark:
                return mem_addr
        return None

    @property
    def is_loaded(self) -> bool:
        """Test if the ALDB is loaded."""
        if self._status == ALDBStatus.LOADING:
            loaded = self._is_loaded()
            if loaded:
                self._update_status(ALDBStatus.LOADED)
        return self._status == ALDBStatus.LOADED

    @property
    def pending_changes(self):
        """Return pending changes."""
        return self._dirty_records

    @property
    def read_write_mode(self) -> ReadWriteMode:
        """Emit the modem read mode."""
        return self._read_write_mode

    @read_write_mode.setter
    def read_write_mode(self, value: ReadWriteMode):
        """Set the modem read mode."""
        self._read_write_mode = ReadWriteMode(value)

    def clear(self):
        """Clear all records from the ALDB.

        This does not write to the device.
        """
        for _, rec in self.items():
            self._notify_change(rec, force_delete=True)
        self._records = {}
        self._dirty_records = {}

    def clear_pending(self):
        """Remove pending changes."""
        self._dirty_records = {}

    def get(self, mem_addr, default=None):
        """Get the record at address 'mem_addr'."""
        return self._records.get(mem_addr, default)

    def get_responders(self, group):
        """Return an iterable of responders to this device for a group.

        Note this only returns the responders known to this device.
        There may be other responders since a device may have
        a `responder` record that is not in the `controller`
        All-Link Database.
        """
        for _, rec in self.items():
            if rec.is_controller and rec.group == group:
                yield rec.target

    def update_version(self, version: EngineVersion):
        """Update the ALDB version number."""
        self._version = EngineVersion(version)

    def subscribe_status_changed(self, listener):
        """Subscribe to notification of ALDB load status changes."""
        subscribe_topic(listener, f"{self._address.id}.{ALDB_STATUS_CHANGED}")

    def subscribe_record_changed(self, listener):
        """Subscribe to notification of ALDB record changes."""
        subscribe_topic(listener, f"{self._address.id}.{ALDB_LINK_CHANGED}")

    def unsubscribe_status_changed(self, listener):
        """Unsubscribe to notification of ALDB load status changes."""
        unsubscribe_topic(listener, f"{self._address.id}.{ALDB_STATUS_CHANGED}")

    def unsubscribe_record_changed(self, listener):
        """Unsubscribe to notification of ALDB record changes."""
        unsubscribe_topic(listener, f"{self._address.id}.{ALDB_LINK_CHANGED}")

    @abstractmethod
    async def async_load(self, *args, **kwargs):
        """Load the All-Link Database."""
        raise NotImplementedError

    def load_saved_records(
        self, status: ALDBStatus, records: List[ALDBRecord], first_mem_addr: int = None
    ):
        """Load All-Link records from a dictionary of saved records."""
        self.clear()
        for mem_addr in records:
            record = records[mem_addr]
            self._records[mem_addr] = record
            self._notify_change(record)

        if self._is_loaded() and self._records:
            self._mem_addr = max(self._records)
            self._update_status(ALDBStatus.LOADED)
            return

        if first_mem_addr is not None:
            self._mem_addr = first_mem_addr

        self._update_status(status)

    def add(
        self,
        group: int,
        target: Address,
        controller: bool = False,
        data1: int = 0x00,
        data2: int = 0x00,
        data3: int = 0x00,
        bit5: int = True,
        bit4: int = False,
    ):
        """Add an All-Link record.

        This method does not write to the device. To write modifications to the device
        use `write` or `async_write`.
        """
        new_rec = ALDBRecord(
            memory=0x0000,
            in_use=True,
            controller=controller,
            high_water_mark=False,
            group=group,
            target=target,
            data1=data1,
            data2=data2,
            data3=data3,
            bit5=bit5,
            bit4=bit4,
        )
        self._add_dirty_record(new_rec)

    def remove(self, mem_addr: int):
        """Remove an All-Link record."""
        rec = self._records.get(mem_addr)
        if not rec:
            raise IndexError("Memory location not found.")

        new_rec = new_aldb_record_from_existing(rec, in_use=False)
        self._add_dirty_record(new_rec)

    def modify(
        self,
        mem_addr: int,
        in_use: bool = None,
        group: int = None,
        controller: bool = None,
        high_water_mark: bool = None,
        target: Address = None,
        data1: int = None,
        data2: int = None,
        data3: int = None,
        bit5: bool = None,
        bit4: bool = None,
    ):
        """Modify an All-Link record."""
        rec = self._records.get(mem_addr)
        if not rec:
            raise IndexError("Memory location not found.")

        new_rec = new_aldb_record_from_existing(
            rec,
            in_use=in_use,
            controller=controller,
            high_water_mark=high_water_mark,
            group=group,
            target=target,
            data1=data1,
            data2=data2,
            data3=data3,
            bit5=bit5,
            bit4=bit4,
        )
        self._add_dirty_record(new_rec)

    async def async_write(self, force=False) -> Tuple[int, int]:
        """Write the dirty records to the device."""
        if not self.is_loaded and not force:
            _LOGGER.warning(
                "ALDB must be loaded before it can be written Status: %s",
                str(self._status),
            )
            return 0, len(self._dirty_records)
        success = 0
        failed = {}
        # make sure to update existing records before adding new ones
        dirty_addrs = sorted(self._dirty_records, reverse=True)
        next_new_dirty = 0
        for dirty_addr in dirty_addrs:
            try:
                rec = self._dirty_records.pop(dirty_addr)
            except KeyError:
                continue
            rec_to_write = rec.copy()
            if rec.mem_addr == 0x0000:
                rec_to_write.mem_addr = self._next_record_mem_addr(
                    target=rec.target,
                    group=rec.group,
                    is_controller=rec.is_controller,
                    data3=rec.data3,
                    force=force,
                )
            result = False
            if self._records.get(rec_to_write.mem_addr) and self._records[
                rec_to_write.mem_addr
            ].is_exact_match(rec_to_write):
                result = ResponseStatus.SUCCESS
            else:
                try:
                    result = await self._write_manager.async_write(rec_to_write, force)
                except ALDBWriteException:
                    result = ResponseStatus.FAILURE

            if result == ResponseStatus.DIRECT_NAK_PRE_NAK and self._read_manager:
                async for test_rec in self._read_manager.async_read(
                    rec_to_write.mem_addr, 1
                ):
                    if rec_to_write.is_exact_match(test_rec):
                        result = ResponseStatus.SUCCESS

            if result == ResponseStatus.SUCCESS:
                curr_rec = self._records.get(rec_to_write.mem_addr)
                # If we wrote to the high water mark, append a new HWM record
                if curr_rec and curr_rec.is_high_water_mark:
                    new_hwm_rec = new_aldb_record_from_existing(
                        HWM_RECORD, mem_addr=curr_rec.mem_addr - 8
                    )
                    self._records[new_hwm_rec.mem_addr] = new_hwm_rec
                self._records[rec_to_write.mem_addr] = rec_to_write
                success += 1
                self._notify_change(rec_to_write)
            else:
                if rec.mem_addr == 0x0000:
                    next_new_dirty -= 1
                    failed[next_new_dirty] = rec
                else:
                    failed[rec.mem_addr] = rec

        self._dirty_records = failed
        return success, len(self._dirty_records)

    def find(
        self,
        group: int = None,
        target: Address = None,
        data3: int = None,
        is_controller: bool = None,
        in_use: bool = None,
    ) -> List[ALDBRecord]:
        """Find all records matching the criteria."""
        if (
            group is None
            and target is None
            and is_controller is None
            and in_use is None
        ):
            raise ValueError("Must have at least one criteria")

        test_rec = ALDBRecord(
            memory=None,
            controller=is_controller,
            group=group,
            target=target,
            data1=None,
            data2=None,
            data3=data3,
            in_use=in_use,
        )
        for _, rec in self.items():
            in_use_match = in_use is None or rec.is_in_use == in_use
            if rec == test_rec and in_use_match:
                yield rec

    def set_load_status(self):
        """Review the ALDB records and identify the load status."""
        _LOGGER.debug("Setting the load status")
        if self._is_loaded():
            self._update_status(ALDBStatus.LOADED)
        elif self._records:
            self._update_status(ALDBStatus.PARTIAL)
        else:
            self._update_status(ALDBStatus.EMPTY)

    def _update_status(self, status):
        """Update the status of the ALDB and notify listeners."""
        new_status = ALDBStatus(int(status))
        if new_status != self._status:
            self._status = new_status
            publish_topic(
                f"{self._address.id}.{ALDB_STATUS_CHANGED}", status=self._status
            )

    def _notify_change(self, record, force_delete=False):
        deleted = True if force_delete else not record.is_in_use
        topic = f"{self._address.id}.{ALDB_LINK_CHANGED}"
        publish_topic(topic, record=record, sender=self.address, deleted=deleted)

    def _next_new_mem_addr(self):
        """Return the next temporary memory address to use for a new record.

        This is a temporary address assigned to a new record until the time
        of writing. When `async_write` is called, an actual address will be
        assigned.
        """
        if not self._dirty_records:
            return -1
        return min(list(self._dirty_records) + [0]) - 1

    def _next_record_mem_addr(
        self,
        target: Address,
        group: int,
        is_controller: bool,
        data3: int,
        force: bool = False,
    ):
        """Assign a memory address to a record.

        Looks for an existing memory address with the same:
        1. Target
        2. Group
        3. Link mode (ie. controller / responder)

        If it finds an existing record with the same values, it overwrites the original rather than
        creating a new record. This ensures there is only one record with the same target, group and link mode.

        If it cannot find an existing record than it returns the next unused record.

        If the ALDB is not loaded it returns an ALDBWriteException
        """
        if not self.is_loaded and not force:
            raise ALDBWriteException("Cannot calculate the next record to write to.")

        for existing_record in self.find(
            target=target,
            group=group,
            is_controller=is_controller,
            data3=data3,
        ):
            return existing_record.mem_addr

        next_mem_addr = self._mem_addr
        for mem_addr, rec in self.items():
            if not rec.is_in_use or rec.is_high_water_mark:
                # This should always be the return if the ALDB is loaded
                return mem_addr
            if mem_addr != next_mem_addr:
                return next_mem_addr
            next_mem_addr = mem_addr - 8
        if not force:
            raise ALDBWriteException(
                "An unknown error in finding the next ALDB record memory address."
            )
        return next_mem_addr

    def _is_loaded(self):
        """Test if the ALDB is fully loaded."""
        has_first = False
        has_last = False
        has_all = True
        prev_addr = 0x0000
        for mem_addr in sorted(self._records, reverse=True):
            if mem_addr == self._mem_addr:
                has_first = True
            if self._records[mem_addr].is_high_water_mark:
                has_last = True
            if prev_addr != 0x0000 and has_all:
                has_all = (prev_addr - mem_addr) == 8
            if len(self._records) == 1 and has_last and has_first:
                # Empty ALDB; yes it is possible with some devices like motion
                has_all = True
            prev_addr = mem_addr
        _LOGGER.debug("Has First is %s", has_first)
        _LOGGER.debug("Has Last is %s", has_last)
        _LOGGER.debug("Has All is %s", has_all)
        return has_first and has_all and has_last

    def _confirm_hwm(self, rec):
        """Confirm the new record is not below the High Water Mark.

        The ALDB will often respond with records that are below the HWM.
        This method confirms no records below the HWM are added.
        If a new HWM is received, this will also remove records below it.

        Records below the HWM are never used by the device and are therefore
        irrelivant.
        """
        curr_hwm_mem_addr = 0x0000
        for curr_mem_addr, curr_rec in self.items():
            if curr_rec.is_high_water_mark:
                curr_hwm_mem_addr = curr_mem_addr
                break

        found_hwm = (
            curr_hwm_mem_addr != 0x0000
            and self._records[curr_hwm_mem_addr].is_high_water_mark
        )

        if rec.is_high_water_mark and rec.mem_addr < curr_hwm_mem_addr:
            curr_hwm_mem_addr = rec.mem_addr

        elif found_hwm and rec.is_high_water_mark and rec.mem_addr > curr_hwm_mem_addr:
            return False

        remove_records = []
        for curr_mem_addr in self._records:
            if found_hwm and curr_mem_addr < curr_hwm_mem_addr:
                remove_records.append(curr_mem_addr)

        for mem_addr in remove_records:
            self._records.pop(mem_addr)

        return True

    def _add_dirty_record(self, mod_rec: ALDBRecord):
        """Add a record to dirty records.

        If the record is FULLY equal to the existing record - Do nothing
        If the record is equal to the existing record with mods - Add with same mem_addr
        If the record is new - Add with new mem_addr
        If the record has existing dirty record - Current record wins
        """

        if mod_rec.mem_addr != 0x0000:
            # Record has a memory address. Assume the user knows what they are doing.
            self._dirty_records[mod_rec.mem_addr] = mod_rec
            return

        for mem_addr in list(self._dirty_records):
            # If there is a functionally same record in the dirty records remove it
            rec = self._dirty_records[mem_addr]
            if rec == mod_rec:
                self._dirty_records.pop(mem_addr)

        data3 = mod_rec.data3 if mod_rec.is_responder else None
        for rec in self.find(
            group=mod_rec.group,
            target=mod_rec.target,
            data3=data3,
            is_controller=mod_rec.is_controller,
            in_use=True,
        ):
            if rec.is_exact_match(mod_rec):
                # No reason to do anything
                return
            # Found a functionally same record (i.e. rec == mod_rec)
            mod_rec.mem_addr = rec.mem_addr

        if mod_rec.mem_addr == 0x0000:
            # This is a new record
            mem_addr = self._next_new_mem_addr()
        else:
            mem_addr = mod_rec.mem_addr

        self._dirty_records[mem_addr] = mod_rec



"""Handle sending a read request for ALDB records."""
import logging

from ...address import Address
from ...constants import ResponseStatus
from ...topics import SET_OPERATING_FLAGS
from .direct_command import DirectCommandHandlerBase

_LOGGER = logging.getLogger(__name__)


class SetOperatingFlagsCommand(DirectCommandHandlerBase):
    """Handle sending a Set Operating Flags command."""

    def __init__(self, address: Address):
        """Init the SetOperatingFlagsCommand."""
        super().__init__(topic=SET_OPERATING_FLAGS, address=address)

    # pylint: disable=arguments-differ
    async def async_send(self, cmd: int, extended=False):
        """Send Get Operating Flags message asyncronously."""
        cmd_response = await super().async_send(cmd=cmd, extended=extended)
        if cmd_response == ResponseStatus.DIRECT_NAK_PRE_NAK and not extended:
            _LOGGER.debug("Attempting resend with extended message")
            return await super().async_send(cmd=cmd, extended=True)
        return cmd_response

    def _update_subscribers_on_direct_ack(
        self, cmd1, cmd2, target, user_data, hops_left
    ):
        """Update subscribers on DIIRECT ack received."""
        self._call_subscribers(response=0)

    def _update_subscribers_on_nak(self, cmd1, cmd2, target, user_data, hops_left):
        """Update subscribers on DIIRECT NAK received."""
        self._call_subscribers(response=cmd2)

    def _update_subscribers_on_direct_nak(
        self, cmd1, cmd2, target, user_data, hops_left
    ):
        """Update subscribers on DIIRECT NAK received."""
        self._call_subscribers(response=cmd2)



"""Manage outbound ON command to a device."""
from ...topics import PRODUCT_DATA_REQUEST
from .direct_command import DirectCommandHandlerBase


class ProductDataRequestCommand(DirectCommandHandlerBase):
    """Manage an outbound ON command to a device."""

    def __init__(self, address):
        """Init the OnLevelCommand class."""
        super().__init__(topic=PRODUCT_DATA_REQUEST, address=address)

    # pylint: disable=arguments-differ
    async def async_send(self):
        """Send the OFF command async."""
        return await super().async_send()

    def _update_subscribers_on_direct_ack(
        self, cmd1, cmd2, target, user_data, hops_left
    ):
        """Update subscribers."""
        if user_data is None:
            return
        product_bytes = bytearray([user_data["d2"], user_data["d3"], user_data["d4"]])
        product_id = int.from_bytes(product_bytes, byteorder="big")
        cat = user_data["d5"]
        subcat = user_data["d6"]
        self._call_subscribers(product_id=product_id, cat=cat, subcat=subcat)



"""Test the Address class."""
import unittest
from binascii import unhexlify

from pyinsteon.address import Address
from tests import set_log_levels


class TestAddress(unittest.TestCase):
    """Test the address class."""

    def setUp(self):
        """Set up the tests."""
        self.hex = "010203"
        self.address = Address(self.hex)
        self.address_bytes = Address(bytearray(unhexlify(self.hex)))
        set_log_levels(
            logger="info",
            logger_pyinsteon="info",
            logger_messages="info",
            logger_topics=False,
        )

    def test_bytes(self):
        """Test address to bytes."""
        assert bytes(self.address) == unhexlify(self.hex)

    def test_high_mid_low(self):
        """Test high, middle and low bytes."""
        assert self.address[0] == 0x01
        assert self.address[1] == 0x02
        assert self.address[2] == 0x03
        try:
            # pylint: disable=unused-variable
            failtest = self.address[3]  # noqa: F841
            assert False
        except ValueError:
            assert True

    def test_eq(self):
        """Test equals."""
        assert self.address == Address("010203")

    def test_ne(self):
        """Test not equals."""
        assert self.address != Address("010204")

    def test_from_byte_array(self):
        """Test create from byte array."""
        assert self.address == self.address_bytes


if __name__ == "__main__":
    unittest.main()



"""Mock reader used by the MockTransport class."""
import asyncio

from aiohttp import web

input_buffer = asyncio.Queue()


def _set_response(web_response: web.Response):
    web_response.headers.add("Content-type", "text/html")
    web_response.set_status(200)
    return web_response


class MockReader:
    """Web server to accept inbound commands."""

    def __init__(self, host="127.0.0.1", port=8080):
        """Init the MockReader class."""
        self._host = host
        self._port = port
        self._app = web.Application()
        self._app.router.add_get("/", self.do_get)
        self._app.router.add_post("/input", self.do_post)
        self._site = None
        self._queue = asyncio.Queue()

    @property
    def queue(self):
        """Return the data queue."""
        return self._queue

    async def async_start(self):
        """Start the web server."""
        runner = web.AppRunner(self._app)
        await runner.setup()
        self._site = web.TCPSite(runner=runner, host=self._host, port=self._port)
        await self._site.start()

    def stop(self):
        """Stop the listener."""
        self._site.stop()

    def do_get(self, request):
        """Respond to standard get requests."""
        response = "<html><head><title>https://pythonbasics.org</title></head>"
        response = f"{response}<body>"
        response = f"{response}<p>Web server is up and listening.  POST messages to `/input`.</p>"
        response = f"{response}</body></html>"
        web_response = web.Response(text=response)
        web_response = _set_response(web_response)
        return web_response

    async def do_post(self, request: web.Request):
        """Handle a post request."""
        json_data = await request.json()
        for _, value in json_data.items():
            await self._queue.put(value)
            await asyncio.sleep(0.5)
        return web.Response(text=str(json_data))



"""Collection of topics mapped to commands (cmd1, cmd2)."""

from collections import namedtuple
import logging
from typing import Tuple

from .topics import (
    ALL_LINK_CLEANUP_STATUS_REPORT,
    ASSIGN_TO_ALL_LINK_GROUP,
    BEEP,
    BRIGHTEN_ONE_STEP,
    DELETE_FROM_ALL_LINK_GROUP,
    DEVICE_TEXT_STRING_REQUEST,
    DIM_ONE_STEP,
    DOOR_CONTROL,
    DOOR_STATUS_REPORT,
    ENTER_LINKING_MODE,
    ENTER_UNLINKING_MODE,
    EXTENDED_GET_RESPONSE,
    EXTENDED_GET_SET,
    EXTENDED_READ_WRITE_ALDB,
    EXTENDED_RECEIVED,
    EXTENDED_TRIGGER_ALL_LINK,
    FACTORY_RESET,
    FX_USERNAME,
    GET_INSTEON_ENGINE_VERSION,
    GET_OPERATING_FLAGS,
    GROUP_OFF,
    GROUP_ON,
    ID_REQUEST,
    INSTANT_CHANGE,
    IO_ALARM_DATA_REQUEST,
    IO_ALARM_DATA_RESPONSE,
    IO_GET_SENSOR_ALARM_DELTA,
    IO_GET_SENSOR_VALUE,
    IO_MODULE_CONTROL,
    IO_OUTPUT_OFF,
    IO_OUTPUT_ON,
    IO_READ_CONFIGURATION_PORT,
    IO_READ_INPUT_PORT,
    IO_SET_SENSOR_1_NOMINAL_VALUE,
    IO_SET_SENSOR_NOMINAL_VALUE,
    IO_WRITE_CONFIGURATION_PORT,
    IO_WRITE_OUTPUT_PORT,
    LEAK_DETECTOR_ANNOUNCE,
    MANUALLY_TURNED_OFF,
    MANUALLY_TURNED_ON,
    NIGHT_MODE_OFF,
    NIGHT_MODE_ON,
    OFF,
    OFF_AT_RAMP_RATE,
    OFF_AT_RAMP_RATE_INBOUND,
    OFF_FAST,
    ON,
    ON_AT_RAMP_RATE,
    ON_AT_RAMP_RATE_INBOUND,
    ON_FAST,
    PEEK,
    PEEK_INTERNAL,
    PING,
    POKE,
    POKE_INTERNAL,
    POOL_CONTROL,
    POOL_DEVICE_OFF,
    POOL_DEVICE_ON,
    POOL_SET_DEVICE_HYSTERESIS,
    POOL_SET_DEVICE_TEMPERATURE,
    POOL_TEMPERATURE_DOWN,
    POOL_TEMPERATURE_UP,
    PRODUCT_DATA_REQUEST,
    REMOTE_SET_BUTTON_TAP1_TAP,
    REMOTE_SET_BUTTON_TAP2_TAP,
    SEND_EXTENDED,
    SEND_STANDARD,
    SET_ADDRESS_MSB,
    SET_ALL_LINK,
    SET_ALL_LINK_COMMAND_ALIAS,
    SET_DEVICE_TEXT_STRING,
    SET_OPERATING_FLAGS,
    SET_SPRINKLER_PROGRAM,
    SET_STATUS,
    SPRINKLER_CONTROL,
    SPRINKLER_GET_PROGRAM_REQUEST,
    SPRINKLER_GET_PROGRAM_RESPONSE,
    SPRINKLER_PROGRAM_OFF,
    SPRINKLER_PROGRAM_ON,
    SPRINKLER_VALVE_OFF,
    SPRINKLER_VALVE_ON,
    STANDARD_RECEIVED,
    START_MANUAL_CHANGE_DOWN,
    START_MANUAL_CHANGE_UP,
    STATUS_REQUEST,
    STOP_MANUAL_CHANGE,
    THERMOSTAT_CONTROL,
    THERMOSTAT_COOL_SET_POINT_STATUS,
    THERMOSTAT_GET_ZONE_INFORMATION,
    THERMOSTAT_HEAT_SET_POINT_STATUS,
    THERMOSTAT_HUMIDITY_STATUS,
    THERMOSTAT_MODE_STATUS,
    THERMOSTAT_SET_COOL_SETPOINT,
    THERMOSTAT_SET_HEAT_SETPOINT,
    THERMOSTAT_TEMPERATURE_DOWN,
    THERMOSTAT_TEMPERATURE_STATUS,
    THERMOSTAT_TEMPERATURE_UP,
    WINDOW_COVERING_CONTROL,
    WINDOW_COVERING_POSITION,
)

Command = namedtuple("Command", "cmd1 cmd2 ud_allowed ud_required userdata")
_LOGGER = logging.getLogger(__name__)


def _check_userdata_match(command: Command, userdata: dict):
    """Check if the command userdata matches the input userdata."""
    if command.ud_allowed and not command.ud_required:
        return True

    if not userdata and not command.ud_required:
        return True

    if userdata and not command.ud_allowed:
        return False

    if command.ud_required and not userdata:
        return False

    for field, value in command.userdata.items():
        if value != userdata.get(field):
            return False
    return True


def _check_match(command, cmd1, cmd2, userdata):
    """Check if the current command matches the input values."""
    if command.cmd1 != cmd1:
        return False
    cmd2_match = cmd2 is None or command.cmd2 is None or command.cmd2 == cmd2
    userdata_match = _check_userdata_match(command, userdata)
    return cmd2_match and userdata_match


class Commands:
    """List of topics and commands."""

    def __init__(self):
        """Init the Commands class."""
        self._topics = {}
        self._commands = {}
        self._use_group = {}
        self._commands_topic_map = {}

    def add(
        self,
        topic: str,
        cmd1: int,
        cmd2: int = None,
        ud_allowed: bool = False,
        ud_required: bool = False,
        userdata: dict = None,
        use_group: bool = False,
    ):
        """Add a command to the list.

        Inputs:
          topic: string
          cmd1: hex byte value of the command
          cmd2: (Optional) hex byte of a specific cmd2 value
          ud_allowed: (Optional, default False)
            - If True, the message is allowed to have user data
            - If False, the messager cannot have user data
          ud_required: (Optional, default False)  Is user data required
            - NOTE: A DIRECT ACK command does not reply with userdata therefore most commands do not require userdata
          userdata: (Optional)  Dictionary of required values in userdata
            - Example: {"d1": 0x00}  Data 1 must be 0x00
        """
        if userdata is not None:
            ud_allowed = True
        if ud_required and userdata is None:
            userdata = {}
        self._topics[topic] = Command(cmd1, cmd2, ud_allowed, ud_required, userdata)
        self._use_group[topic] = use_group
        if self._commands_topic_map.get(cmd1) is None:
            self._commands_topic_map[cmd1] = []
        self._commands_topic_map[cmd1].append(topic)

    def get(self, topic: str) -> Command:
        """Get the command elements of the topic."""
        return self._topics.get(topic)

    def get_command(self, topic: str) -> Tuple[int, int, bool]:
        """Get cmd1 and cmd2 from a topic.

        Returns (cmd1, cmd2, extended)
        """
        return self._topics.get(topic)

    def _get_topics_from_cmd1(self, cmd1):
        """Get a list of topics from a cmd1 value."""
        return self._commands_topic_map.get(cmd1)

    def use_group(self, topic):
        """Return if a topic requires a group number."""
        return self._use_group.get(topic)

    def get_topics(self, cmd1, cmd2, flags, userdata=None, send=False):
        """Generate a topic from a cmd1, cmd2 and extended flag."""
        found = False
        for topic in self._commands_topic_map.get(cmd1, {}):
            command = self._topics[topic]
            if _check_match(command, cmd1, cmd2, userdata):
                found = True
                yield topic
            elif flags.is_direct_nak and _check_match(command, cmd1, None, userdata):
                found = True
                yield topic
        if not found:
            if userdata is None:
                yield SEND_STANDARD if send else STANDARD_RECEIVED
            else:
                yield SEND_EXTENDED if send else EXTENDED_RECEIVED


commands = Commands()

commands.add(
    topic=STANDARD_RECEIVED,
    cmd1=-1,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=EXTENDED_RECEIVED,
    cmd1=-1,
    cmd2=None,
    ud_allowed=True,
    ud_required=True,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=SEND_STANDARD,
    cmd1=-2,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=SEND_EXTENDED,
    cmd1=-2,
    cmd2=None,
    ud_allowed=True,
    ud_required=True,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=ASSIGN_TO_ALL_LINK_GROUP,
    cmd1=0x01,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=DELETE_FROM_ALL_LINK_GROUP,
    cmd1=0x02,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=PRODUCT_DATA_REQUEST,
    cmd1=0x03,
    cmd2=0x00,
    ud_allowed=True,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=FX_USERNAME,
    cmd1=0x03,
    cmd2=0x01,
    ud_allowed=True,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=DEVICE_TEXT_STRING_REQUEST,
    cmd1=0x03,
    cmd2=0x02,
    ud_allowed=True,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=SET_DEVICE_TEXT_STRING,
    cmd1=0x03,
    cmd2=0x03,
    ud_allowed=True,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=SET_ALL_LINK_COMMAND_ALIAS,
    cmd1=0x03,
    cmd2=0x04,
    ud_allowed=True,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=SET_ALL_LINK,
    cmd1=0x03,
    cmd2=0x05,
    ud_allowed=True,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=ALL_LINK_CLEANUP_STATUS_REPORT,
    cmd1=0x06,
    cmd2=None,
    ud_allowed=True,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=ENTER_LINKING_MODE,
    cmd1=0x09,
    cmd2=None,
    ud_allowed=True,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=ENTER_UNLINKING_MODE,
    cmd1=0x0A,
    cmd2=None,
    ud_allowed=True,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=GET_INSTEON_ENGINE_VERSION,
    cmd1=0x0D,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=PING,
    cmd1=0x0F,
    cmd2=0x00,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=ID_REQUEST,
    cmd1=0x10,
    cmd2=0x00,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=ON,
    cmd1=0x11,
    cmd2=None,
    ud_allowed=True,
    ud_required=False,
    userdata=None,
    use_group=True,
)
commands.add(
    topic=ON_FAST,
    cmd1=0x12,
    cmd2=None,
    ud_allowed=True,
    ud_required=False,
    userdata=None,
    use_group=True,
)
commands.add(
    topic=OFF,
    cmd1=0x13,
    cmd2=None,
    ud_allowed=True,
    ud_required=False,
    userdata=None,
    use_group=True,
)
commands.add(
    topic=OFF_FAST,
    cmd1=0x14,
    cmd2=None,
    ud_allowed=True,
    ud_required=False,
    userdata=None,
    use_group=True,
)
commands.add(
    topic=BRIGHTEN_ONE_STEP,
    cmd1=0x15,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=True,
)
commands.add(
    topic=DIM_ONE_STEP,
    cmd1=0x16,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=True,
)
commands.add(
    topic=START_MANUAL_CHANGE_DOWN,
    cmd1=0x17,
    cmd2=0x00,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=True,
)
commands.add(
    topic=START_MANUAL_CHANGE_UP,
    cmd1=0x17,
    cmd2=0x01,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=True,
)
commands.add(
    topic=STOP_MANUAL_CHANGE,
    cmd1=0x18,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=True,
)
commands.add(
    topic=STATUS_REQUEST,
    cmd1=0x19,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=GET_OPERATING_FLAGS,
    cmd1=0x1F,
    cmd2=None,
    ud_allowed=True,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=SET_OPERATING_FLAGS,
    cmd1=0x20,
    cmd2=None,
    ud_allowed=True,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=INSTANT_CHANGE,
    cmd1=0x21,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=True,
)
commands.add(
    topic=MANUALLY_TURNED_OFF,
    cmd1=0x22,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=True,
)
commands.add(
    topic=MANUALLY_TURNED_ON,
    cmd1=0x23,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=True,
)
commands.add(
    topic=REMOTE_SET_BUTTON_TAP1_TAP,
    cmd1=0x25,
    cmd2=0x01,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=REMOTE_SET_BUTTON_TAP2_TAP,
    cmd1=0x25,
    cmd2=0x02,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=SET_STATUS,
    cmd1=0x27,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=SET_ADDRESS_MSB,
    cmd1=0x28,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=POKE,
    cmd1=0x29,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=PEEK,
    cmd1=0x2B,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=PEEK_INTERNAL,
    cmd1=0x2C,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=POKE_INTERNAL,
    cmd1=0x2D,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=ON_AT_RAMP_RATE,
    cmd1=0x2E,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=EXTENDED_GET_SET,
    cmd1=0x2E,
    cmd2=None,
    ud_allowed=True,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=EXTENDED_GET_RESPONSE,
    cmd1=0x2E,
    cmd2=None,
    ud_allowed=True,
    ud_required=True,
    userdata=None,
    use_group=False,
)
# cmd2 ne 0x00 => no confict w/ read aldb
commands.add(
    topic=OFF_AT_RAMP_RATE,
    cmd1=0x2F,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
# direct is ed and direct_ack is sd
commands.add(
    topic=EXTENDED_READ_WRITE_ALDB,
    cmd1=0x2F,
    cmd2=None,
    ud_allowed=True,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=EXTENDED_TRIGGER_ALL_LINK,
    cmd1=0x30,
    cmd2=0x00,
    ud_allowed=True,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=BEEP,
    cmd1=0x30,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=GROUP_ON,
    cmd1=0x32,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=GROUP_OFF,
    cmd1=0x33,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=FACTORY_RESET,
    cmd1=0x34,
    cmd2=0x00,
    ud_allowed=True,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=ON_AT_RAMP_RATE_INBOUND,
    cmd1=0x34,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=True,
)
commands.add(
    topic=OFF_AT_RAMP_RATE_INBOUND,
    cmd1=0x35,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=True,
)
commands.add(
    topic=NIGHT_MODE_ON,
    cmd1=0x3B,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=NIGHT_MODE_OFF,
    cmd1=0x3C,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=SET_SPRINKLER_PROGRAM,
    cmd1=0x40,
    cmd2=None,
    ud_allowed=True,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=SPRINKLER_VALVE_ON,
    cmd1=0x40,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=SPRINKLER_GET_PROGRAM_RESPONSE,
    cmd1=0x41,
    cmd2=None,
    ud_allowed=True,
    ud_required=True,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=SPRINKLER_VALVE_OFF,
    cmd1=0x41,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=SPRINKLER_PROGRAM_ON,
    cmd1=0x42,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=SPRINKLER_PROGRAM_OFF,
    cmd1=0x43,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=SPRINKLER_CONTROL,
    cmd1=0x44,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=SPRINKLER_GET_PROGRAM_REQUEST,
    cmd1=0x45,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=IO_OUTPUT_ON,
    cmd1=0x45,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=IO_OUTPUT_OFF,
    cmd1=0x46,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=IO_ALARM_DATA_REQUEST,
    cmd1=0x47,
    cmd2=0x00,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=IO_WRITE_OUTPUT_PORT,
    cmd1=0x48,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=IO_READ_INPUT_PORT,
    cmd1=0x49,
    cmd2=0x00,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=IO_GET_SENSOR_VALUE,
    cmd1=0x4A,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=IO_SET_SENSOR_1_NOMINAL_VALUE,
    cmd1=0x4B,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=IO_SET_SENSOR_NOMINAL_VALUE,
    cmd1=0x4B,
    cmd2=None,
    ud_allowed=True,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=IO_GET_SENSOR_ALARM_DELTA,
    cmd1=0x4C,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=IO_ALARM_DATA_RESPONSE,
    cmd1=0x4C,
    cmd2=0x00,
    ud_allowed=True,
    ud_required=True,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=IO_WRITE_CONFIGURATION_PORT,
    cmd1=0x4D,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=IO_READ_CONFIGURATION_PORT,
    cmd1=0x4E,
    cmd2=0x00,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=IO_MODULE_CONTROL,
    cmd1=0x4F,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=POOL_DEVICE_ON,
    cmd1=0x50,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=POOL_SET_DEVICE_TEMPERATURE,
    cmd1=0x50,
    cmd2=0x00,
    ud_allowed=True,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=POOL_SET_DEVICE_HYSTERESIS,
    cmd1=0x50,
    cmd2=0x01,
    ud_allowed=True,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=POOL_DEVICE_OFF,
    cmd1=0x51,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=POOL_TEMPERATURE_UP,
    cmd1=0x52,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=POOL_TEMPERATURE_DOWN,
    cmd1=0x53,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=POOL_CONTROL,
    cmd1=0x54,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=DOOR_CONTROL,
    cmd1=0x58,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=DOOR_STATUS_REPORT,
    cmd1=0x59,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=WINDOW_COVERING_CONTROL,
    cmd1=0x60,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=WINDOW_COVERING_POSITION,
    cmd1=0x61,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=THERMOSTAT_TEMPERATURE_UP,
    cmd1=0x68,
    cmd2=None,
    ud_allowed=True,
    ud_required=False,
    userdata=True,
    use_group=False,
)
commands.add(
    topic=THERMOSTAT_TEMPERATURE_DOWN,
    cmd1=0x69,
    cmd2=None,
    ud_allowed=True,
    ud_required=False,
    userdata=True,
    use_group=False,
)
commands.add(
    topic=THERMOSTAT_GET_ZONE_INFORMATION,
    cmd1=0x6A,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=THERMOSTAT_CONTROL,
    cmd1=0x6B,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=True,
    use_group=False,
)
commands.add(
    topic=THERMOSTAT_SET_COOL_SETPOINT,
    cmd1=0x6C,
    cmd2=None,
    ud_allowed=True,
    ud_required=False,
    userdata=True,
    use_group=False,
)
commands.add(
    topic=THERMOSTAT_SET_HEAT_SETPOINT,
    cmd1=0x6D,
    cmd2=None,
    ud_allowed=True,
    ud_required=False,
    userdata=True,
    use_group=False,
)
commands.add(
    topic=THERMOSTAT_TEMPERATURE_STATUS,
    cmd1=0x6E,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=THERMOSTAT_HUMIDITY_STATUS,
    cmd1=0x6F,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=LEAK_DETECTOR_ANNOUNCE,
    cmd1=0x70,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=THERMOSTAT_MODE_STATUS,
    cmd1=0x70,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=THERMOSTAT_COOL_SET_POINT_STATUS,
    cmd1=0x71,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)
commands.add(
    topic=THERMOSTAT_HEAT_SET_POINT_STATUS,
    cmd1=0x72,
    cmd2=None,
    ud_allowed=False,
    ud_required=False,
    userdata=None,
    use_group=False,
)



"""Insteon inbound message data structure defintion."""
from binascii import hexlify
import logging
from typing import Tuple

from . import MessageBase
from ...constants import AckNak, MessageId
from ...data_types.message_flags import MessageFlags
from .message_definition import MessageDefinition
from .message_definitions import FLD_EXT_SEND_ACK, INBOUND_MSG_DEF, MessageField

_LOGGER = logging.getLogger(__name__)


def trim_data(raw_data: bytearray):
    """Trim bad data from the front of a message byte stream."""
    data_bytes = bytes(raw_data)
    while data_bytes and data_bytes[0] != 0x02:
        data_bytes = data_bytes[1:]
    return data_bytes


class Inbound(MessageBase):
    """Insteon inbound message data definition."""

    def __init__(self, msg_def: MessageDefinition, raw_data: bytearray):
        """Init the Inbound message class."""
        self._fields = msg_def.fields
        self._len = len(msg_def)
        field_vals = self._slice_data(msg_def.slices, raw_data)
        super().__init__(msg_def, **field_vals)

    def __len__(self):
        """Emit the length of the message."""
        return self._len

    def _slice_data(self, slices, raw_data):
        curr_slice = 2
        field_vals = {}
        for field in self._fields:
            val = field.type(raw_data[slices[curr_slice]])
            field_vals[field.name] = val
            curr_slice += 1
        return field_vals

    def __str__(self):
        """Emit the message in hex."""
        return hexlify(bytes(self)).decode()


def create(raw_data: bytearray) -> Tuple[Inbound, bytearray]:
    """Create a message from a raw byte array."""

    def _remaining_data(msg, data_bytes):
        if msg is not None:
            return data_bytes[len(msg) :]
        return data_bytes

    def _create_message(msg_def, data_bytes):
        msg = Inbound(msg_def, data_bytes)
        return msg, _remaining_data(msg, data_bytes)

    def _standard_message(data_bytes):
        flag_byte = 5
        flags = MessageFlags(data_bytes[flag_byte])
        if flags.is_extended:
            msg_def = MessageDefinition(MessageId.SEND_EXTENDED, FLD_EXT_SEND_ACK)
            if len(data_bytes) < len(msg_def):
                _LOGGER.debug("Full extended message not received")
                _LOGGER.debug("Returning: %s", data_bytes.hex())
                return None, bytearray(data_bytes)
        else:
            msg_def = INBOUND_MSG_DEF.get(MessageId.SEND_STANDARD)
        msg, remaining_data = _create_message(msg_def, data_bytes)
        return msg, remaining_data

    _LOGGER.debug("IN CREATE: %s", raw_data.hex())
    data_bytes = trim_data(raw_data)
    if len(data_bytes) < 2:
        _LOGGER.debug("Message less than 2 bytes")
        _LOGGER.debug("Returning: %s", data_bytes.hex())
        return None, bytearray(data_bytes)
    try:
        msg_id = MessageId(data_bytes[1])
    except ValueError as err:
        _LOGGER.debug("Error: %s", err)
        truncate = 1 if data_bytes[1] == 0x02 else 2
        data_bytes = trim_data(bytearray(data_bytes[truncate:]))
        _LOGGER.debug("Returning: %s", data_bytes.hex())
        return None, bytearray(data_bytes)
    if (
        msg_id in [MessageId.GET_IM_CONFIGURATION, MessageId.GET_IM_INFO]
        and len(data_bytes) >= 3
        and data_bytes[2] == 0x15
    ):
        msg_def = MessageDefinition(msg_id, [MessageField("ack", 1, AckNak)])
    else:
        msg_def = INBOUND_MSG_DEF.get(msg_id)

    if msg_def is not None:
        if len(data_bytes) < len(msg_def):
            _LOGGER.debug("Full message not received")
            _LOGGER.debug("Returning: %s", data_bytes.hex())
            return None, bytearray(data_bytes)

        if msg_id == MessageId.SEND_STANDARD:
            msg, remaining_data = _standard_message(data_bytes)
        else:
            msg, remaining_data = _create_message(msg_def, data_bytes)
        _LOGGER.debug("Returning: %s", remaining_data.hex())
        return msg, bytearray(remaining_data)

    return None, trim_data(data_bytes[1:])



"""Receive an All-Link record for an Insteon Modem."""

import logging

from ..address import Address
from ..constants import AllLinkMode
from ..topics import ALL_LINKING_COMPLETED
from . import inbound_handler
from .inbound_base import InboundHandlerBase

_LOGGER = logging.getLogger(__name__)


class AllLinkCompletedHandler(InboundHandlerBase):
    """Receive an All-Link Completed message for an Insteon Modem."""

    def __init__(self):
        """Init the AllLinkRecordResponse class."""
        super().__init__(ALL_LINKING_COMPLETED)

    @inbound_handler
    def handle_response(
        self,
        link_mode: AllLinkMode,
        group: int,
        target: Address,
        cat: int,
        subcat: int,
        firmware: int,
    ):
        """Recieve an all link record."""
        self._call_subscribers(
            link_mode=link_mode,
            group=group,
            target=target,
            cat=cat,
            subcat=subcat,
            firmware=firmware,
        )



"""Energy Management Control devices (CATEGORY 0x09)."""

from ..config import LOAD_SENSE_ON, PROGRAM_LOCK_ON
from ..events import OFF_EVENT, OFF_FAST_EVENT, ON_EVENT, ON_FAST_EVENT
from ..groups import LOAD_SENSOR, ON_OFF_SWITCH
from .device_commands import STATUS_COMMAND
from .on_off_responder_base import OnOffResponderBase


class EnergyManagement_LoadController(OnOffResponderBase):
    """Load Controller device."""

    def __init__(
        self,
        address,
        cat,
        subcat,
        firmware=0x00,
        description="",
        model="",
        buttons=None,
        state_name=ON_OFF_SWITCH,
        on_event_name=ON_EVENT,
        off_event_name=OFF_EVENT,
        on_fast_event_name=ON_FAST_EVENT,
        off_fast_event_name=OFF_FAST_EVENT,
    ):
        """Init the EnergyManagement_LoadController class."""
        buttons = (
            {1: (ON_OFF_SWITCH, 0), 2: (LOAD_SENSOR, 1)} if buttons is None else buttons
        )
        super().__init__(
            address,
            cat,
            subcat,
            firmware,
            description,
            model,
            buttons,
            on_event_name,
            off_event_name,
            on_fast_event_name,
            off_fast_event_name,
        )

    def _register_op_flags_and_props(self):
        super()._register_op_flags_and_props()
        self._add_operating_flag(PROGRAM_LOCK_ON, 0, 0, 0, 1)
        self._add_operating_flag(LOAD_SENSE_ON, 0, 5, 6, 7)

    def _subscribe_to_handelers_and_managers(self):
        super()._subscribe_to_handelers_and_managers()
        self._managers[STATUS_COMMAND].add_status_type(1, self._sensor_status)

    def _sensor_status(self, db_version, status):
        """Set the sensor status value."""
        self._groups[2].set_value(status)
